import os
import logging

import numpy as np
import pytest

import capytaine as cpt
from capytaine.bem.problems_and_results import LinearPotentialFlowProblem, \
        LinearPotentialFlowResult, DiffractionResult, RadiationResult
from capytaine.io.legacy import import_cal_file


#####################################
#  Setup frequency and water_depth  #
#####################################

@pytest.mark.parametrize("omega", [1.0, 2.0])
def test_compute_all_derived_freqs_infinite_depth(omega):
    pb = LinearPotentialFlowProblem(omega=omega, water_depth=np.inf)
    assert pb.omega == omega
    assert pb.freq == omega/(2*np.pi)
    assert pb.period == 2*np.pi/omega
    assert pb.wavenumber == omega**2/9.81
    assert pb.wavelength == 9.81*2*np.pi/omega**2


@pytest.mark.parametrize("omega", [1.0, 2.0])
def test_compute_derived_freqs_finite_depth(omega):
    pb = LinearPotentialFlowProblem(free_surface=0.0, water_depth=1.0, omega=omega)
    assert pb.water_depth == 1.0
    assert np.isclose(pb.omega**2, pb.g*pb.wavenumber*np.tanh(pb.wavenumber*pb.water_depth))


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_wavelength(water_depth):
    λ = 10*np.random.rand()
    assert np.isclose(LinearPotentialFlowProblem(wavelength=λ, water_depth=water_depth).wavelength, λ)


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_wavenumber(water_depth):
    k = 10*np.random.rand()
    assert np.isclose(LinearPotentialFlowProblem(wavenumber=k, water_depth=water_depth).wavenumber, k)


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_period(water_depth):
    T = 10*np.random.rand()
    assert np.isclose(LinearPotentialFlowProblem(period=T, water_depth=water_depth).period, T)


@pytest.mark.parametrize("water_depth", [10.0, np.inf])
def test_setting_freq(water_depth):
    f = 10*np.random.rand()
    assert np.isclose(LinearPotentialFlowProblem(freq=f, water_depth=water_depth).freq, f)


def test_setting_too_many_frequencies():
    with pytest.raises(ValueError, match="at most one"):
        cpt.DiffractionProblem(omega=1.0, wavelength=1.0)


def test_problems_ordering():
    pb1 = LinearPotentialFlowProblem(omega=1.0)
    pb2 = LinearPotentialFlowProblem(omega=2.0)
    assert pb1 < pb2


def test_define_problems_with_invalid_free_surface_value():
    with pytest.raises(NotImplementedError):
        LinearPotentialFlowProblem(free_surface=2.0)


def test_define_problems_without_free_surface_but_with_sea_bottom():
    with pytest.raises(NotImplementedError):
        LinearPotentialFlowProblem(free_surface=np.inf, water_depth=2.0)


def test_define_problems_with_negative_depth():
    with pytest.raises(ValueError):
        LinearPotentialFlowProblem(water_depth=-1.0)


def test_backward_compatibility_with_sea_bottom_argument(caplog):
    with caplog.at_level(logging.WARNING):
        pb = cpt.DiffractionProblem(sea_bottom=-10.0)
        assert pb.water_depth == 10.0
    assert 'water_depth' in caplog.text


def test_define_problem_with_subtype_arguments():
    with pytest.raises(TypeError):
        LinearPotentialFlowProblem(radiating_dof="Heave")


#################################
#  Body and boundary condition  #
#################################

@pytest.fixture
def sphere():
    sphere = cpt.FloatingBody(
            mesh=cpt.mesh_sphere(center=(0, 0, -2), radius=1.0, resolution=(5, 5)),
            name="sphere",
            )
    sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    return sphere


def test_define_problem_without_body():
    assert LinearPotentialFlowProblem().body is None


def test_define_problem_with_empty_body():
    with pytest.raises(ValueError):
        LinearPotentialFlowProblem(body=cpt.FloatingBody(mesh=cpt.Mesh([], [])))


def test_define_generic_problem_with_its_boundary_condition(sphere):
    bc = sphere.mesh.faces_normals @ (1, 1, 1)
    pb = LinearPotentialFlowProblem(body=sphere, boundary_condition=bc)
    np.testing.assert_allclose(pb.boundary_condition, bc)


def test_influenced_dofs(sphere):
    pb = LinearPotentialFlowProblem(body=sphere)
    assert list(pb.influenced_dofs.keys()) == ['Heave']


def test_mesh_inconsistent_with_dofs(sphere):
    n = sphere.mesh.nb_faces
    sphere.dofs["dof_with_wrong_shape"] = np.ones((n//2, 3))
    with pytest.raises(ValueError):
        cpt.RadiationProblem(body=sphere, wavenumber=1.0)
    with pytest.raises(ValueError):
        cpt.DiffractionProblem(body=sphere, wavenumber=1.0)


def test_mesh_above_the_free_surface(caplog):
    with caplog.at_level(logging.WARNING):
        body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
        pb = cpt.RadiationProblem(body=body, omega=1.0)
    assert '50 panels above the free surface' in caplog.text
    assert np.all(pb.body.mesh.vertices[:, 2].max() <= 0.0)


def test_mesh_below_the_sea_bottom(caplog):
    with caplog.at_level(logging.WARNING):
        body = cpt.FloatingBody(
                mesh=cpt.mesh_sphere(radius=0.5, center=(0, 0, -1.0)),
                dofs=cpt.rigid_body_dofs()
                )
        pb = cpt.RadiationProblem(body=body, omega=1.0, water_depth=1.0)
    assert '50 panels below the sea bottom' in caplog.text
    assert np.all(pb.body.mesh.vertices[:, 2].min() >= -1.0)


def test_diffraction_boundary_condition(sphere):
    pb = cpt.DiffractionProblem(body=sphere, wave_direction=1.0, wavenumber=1.0)
    assert len(pb.boundary_condition) == sphere.mesh.nb_faces


def test_cannot_define_boundary_condition_of_diffraction_problem():
    with pytest.raises(TypeError):
        cpt.DiffractionProblem(boundary_conditions=[0, 0, 0])


def test_diffraction_problem_str():
    assert "DiffractionProblem" in str(cpt.DiffractionProblem(wavenumber=1.0, g=10, rho=1025, free_surface=np.inf))


def test_wave_direction_radians_warning(sphere, caplog):
    with caplog.at_level(logging.WARNING):
        cpt.DiffractionProblem(body=sphere, omega=1.0, wave_direction=180)
    assert 'in radians and not in degrees' in caplog.text


def test_radiation_problem_with_wrong_dof_name():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    with pytest.raises(ValueError, match="'Flurp'"):
        cpt.RadiationProblem(body=body, wavelength=1.0, radiating_dof="Flurp")


def test_radiation_problem_boundary_condition(sphere):
    pb = cpt.RadiationProblem(body=sphere, omega=1.0)
    assert len(pb.boundary_condition) == sphere.mesh.nb_faces


def test_radiation_problem_str():
    assert "RadiationProblem" in str(cpt.RadiationProblem(g=10, rho=1025, free_surface=np.inf))


############################
#  Resolution and results  #
############################

def test_transform_into_result(sphere):
    pb = LinearPotentialFlowProblem(body=sphere)
    res = pb.make_results_container()
    assert isinstance(res, LinearPotentialFlowResult)
    assert res.problem is pb
    assert res.omega == pb.omega
    assert res.period == pb.period
    assert res.body is pb.body

def test_transform_into_result_and_set_result_values(sphere):
    pb = LinearPotentialFlowProblem()
    res = pb.make_results_container(forces={"Heave": 1.0 + 2.0j}, potential=np.array([]), sources=np.array([]), pressure=np.array([]))
    assert res.forces["Heave"] == 1.0 + 2.0j


def test_transform_diffraction_problem_into_diffraction_result():
    pb = cpt.DiffractionProblem()
    res = pb.make_results_container()
    assert isinstance(res, DiffractionResult)


def test_inherit_parameters():
    pb = cpt.DiffractionProblem(g=10, rho=1023, free_surface=np.inf)
    res = pb.make_results_container()
    assert res.g == pb.g == 10


def test_diffraction_result_str():
    res = cpt.DiffractionProblem().make_results_container()
    assert "DiffractionResult" in str(res)


def test_radiation_result_str():
    res = cpt.RadiationProblem().make_results_container()
    assert "RadiationResult" in str(res)


def test_transform_radiation_problem_into_radiation_result():
    pb = cpt.RadiationProblem()
    res = pb.make_results_container()
    assert isinstance(res, RadiationResult)
    assert res.added_mass == res.added_masses == {}
    assert res.radiation_damping == res.radiation_dampings == {}


def test_transform_radiation_problem_into_radiation_result_and_setting_force():
    pb = cpt.RadiationProblem()
    res = pb.make_results_container(forces={"Heave": 1.0 + 2.0j})
    assert res.added_mass == res.added_masses == {"Heave": 1.0}
    assert res.radiation_damping == res.radiation_dampings == {"Heave": 2.0}


####################
#  Failed results  #
####################

@pytest.fixture
def broken_bem_solver():
    ref_gf = cpt.Delhommeau()
    class BrokenGreenFunction:
        def evaluate(self, m1, m2, fs, wd, wavenumber, *args, **kwargs):
            if wavenumber < 2.0:
                raise NotImplementedError("I'm potato")
            else:
                return ref_gf.evaluate(m1, m2, fs, wd, wavenumber, *args, **kwargs)
    broken_bem_solver = cpt.BEMSolver(green_function=BrokenGreenFunction())
    return broken_bem_solver


def test_failed_resolution_failing(broken_bem_solver, sphere):
    pb = cpt.DiffractionProblem(body=sphere, wavenumber=1.0, wave_direction=0.0)
    with pytest.raises(NotImplementedError):
        broken_bem_solver.solve(pb)


def test_failed_resolution_catched(broken_bem_solver, sphere):
    from capytaine.bem.problems_and_results import FailedDiffractionResult
    pb = cpt.DiffractionProblem(body=sphere, wavenumber=1.0, wave_direction=0.0)
    failed_res = broken_bem_solver.solve_all([pb])[0]
    assert isinstance(failed_res, FailedDiffractionResult)


def test_failed_resolution_replaced_by_nan(broken_bem_solver, sphere):
    pb = cpt.RadiationProblem(body=sphere, wavenumber=1.0, radiating_dof="Heave")
    failed_ds = cpt.assemble_dataset(broken_bem_solver.solve_all([pb]))
    assert np.isnan(failed_ds.added_mass.values[0, 0, 0])
    assert np.isnan(failed_ds.radiation_damping.values[0, 0, 0])


########
#  IO  #
########

@pytest.mark.parametrize("cal_file", ["Nemoh.cal", "Nemoh_v3.cal"])
def test_import_cal_file(cal_file):
    """Test the importation of legacy Nemoh.cal files."""
    current_file_path = os.path.dirname(os.path.abspath(__file__))

    # Non symmetrical body
    cal_file_path = os.path.join(current_file_path, "Nemoh_verification_cases", "NonSymmetrical", cal_file)
    problems = import_cal_file(cal_file_path)

    assert len(problems) == 6*41+41
    for problem in problems:
        assert problem.rho == 1000.0
        assert problem.g == 9.81
        assert problem.water_depth == np.inf
        assert isinstance(problem.body, cpt.FloatingBody)
        assert problem.body.nb_dofs == 6
        assert problem.body.mesh.nb_vertices == 299  # Duplicate vertices are removed during import.
        assert problem.body.mesh.nb_faces == 280
        assert problem.omega in np.linspace(0.1, 2.0, 41)
        if isinstance(problem, cpt.DiffractionProblem):
            assert problem.wave_direction == 0.0

    # Symmetrical cylinder
    cal_file_path = os.path.join(current_file_path, "Nemoh_verification_cases", "Cylinder", cal_file)
    problems = import_cal_file(cal_file_path)

    assert len(problems) == 6*2+2
    for problem in problems:
        assert problem.rho == 1000.0
        assert problem.g == 9.81
        assert problem.water_depth == np.inf
        assert isinstance(problem.body.mesh, cpt.ReflectionSymmetricMesh)
        assert isinstance(problem.body.mesh[0], cpt.Mesh)
        assert problem.body.nb_dofs == 6
        # assert problem.body.mesh.nb_vertices == 2*540
        assert problem.body.mesh.nb_faces == 2*300
        assert problem.omega == 0.1 or problem.omega == 2.0
        if isinstance(problem, cpt.DiffractionProblem):
            assert problem.wave_direction == 0.0



import numpy as np
import xarray as xr
import capytaine as cpt

mesh_1 = cpt.load_mesh("./boat_200.mar", file_format="nemoh")
body_1 = cpt.FloatingBody(
            mesh=mesh_1,
            lid_mesh=mesh_1.generate_lid(faces_max_radius=1.0),  # Adjust to the desired mesh resolution or remove lid entirely
            dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)),
            center_of_mass=(0, 0, 0)  # Optional, only for hydrostatics
        )
body_1.inertia_matrix = body_1.compute_rigid_body_inertia(rho=1025)
body_1.hydrostatic_stiffness = body_1.immersed_part().compute_hydrostatic_stiffness(rho=1025)

# If you have several rigid bodies, copy the code above to define "body_2", "body_3", etc.

all_bodies = body_1  # Replace "body_1" by "body_1 + body_2 + body_3" for multibody problem.
all_bodies = all_bodies.immersed_part()  # if the mesh has panels above the free surface, this command removes them

# Set up parameters
test_matrix = xr.Dataset({
        "omega": np.linspace(0.1, 2.0, 20),  # Can also specify "freq" (in Hz), "period", "wavelength" or "wavenumber"
        "wave_direction": np.linspace(0, np.pi, 3),
        "radiating_dof": list(all_bodies.dofs),
        "water_depth": [np.inf],
        "rho": [1025],
        })

# Do the resolution
solver = cpt.BEMSolver()
dataset = solver.fill_dataset(test_matrix, all_bodies)

# Export to netcdf file
cpt.export_dataset("capy_dataset.nc", dataset)



import pytest
import numpy as np

import capytaine as cpt
from capytaine.bem.airy_waves import froude_krylov_force

from scipy.special import j0, y0, jn, yn  # Bessel functions


@pytest.mark.parametrize("k", [5.0, 10.0])
@pytest.mark.parametrize("gf", [cpt.Delhommeau(), cpt.FinGreen3D()])
@pytest.mark.parametrize("method", ["direct", "indirect"])
def test_mccamy_and_fuchs(k, gf, method):
    h = 1.0  # Water depth
    R = 0.3  # Cylinder radius

    mesh = cpt.mesh_vertical_cylinder(radius=R, length=h, center=(0, 0, -h/2), resolution=(0, 30, 20)).immersed_part(water_depth=h)
    body = cpt.FloatingBody(mesh=mesh, dofs={"Surge": np.array([[1.0, 0.0, 0.0] for face in mesh.faces])})
    pb = cpt.DiffractionProblem(body=body, wavenumber=k, wave_direction=0.0, water_depth=h)

    # Analytical solution of McCamy and Fuchs for wavenumber k
    m0 = pb.omega**2/pb.g
    dy = (y0(m0*R) - yn(2, m0*R))
    dj = (j0(m0*R) - jn(2, m0*R))
    analytical_force = (4*pb.rho*pb.g*h*R) * (-1j*2*h/R) * np.tanh(m0*h)/(m0*h)**2 * (dy + 1j*dj)/(dy**2 + dj**2)

    solver = cpt.BEMSolver(method=method, green_function=gf)
    res = solver.solve(pb)
    numerical_force = res.force["Surge"] + froude_krylov_force(res)["Surge"]

    np.testing.assert_allclose(numerical_force, analytical_force, rtol=0.05)



"""Prony decomposition: tool to approximate a function as a sum of exponentials.
Used in particular in the finite depth Green function.
"""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging

import numpy as np
from numpy.polynomial import polynomial
from scipy.optimize import curve_fit
from scipy.linalg import toeplitz

LOG = logging.getLogger(__name__)
RNG = np.random.default_rng()


def exponential_decomposition(X, F, m):
    """Use Prony's method to approximate the sampled real function F=f(X) as a sum of m
    exponential functions x → Σ a_i exp(lamda_i x).

    Parameters
    ----------
    X: 1D array
        sampling points.
    F: 1D array (same size as X)
        values of the function to approximate at the points of x.
    m: integer
        number of exponential functions

    Return
    ------
    a: 1D array (size m)
        coefficients of the exponentials
    lamda: 1D array (size m)
        growth rate of the exponentials
    """
    assert X.shape == F.shape

    # Compute the coefficients of the polynomials of Prony's method
    A = toeplitz(c=F[m-1:-1], r=F[:m][::-1])
    P, *_ = np.linalg.lstsq(A, F[m:], rcond=None)

    # Build and solve polynomial function
    coeffs = np.ones(m+1)
    # coeffs[:m] = -P[::-1]
    for i in range(m):
        coeffs[m-i-1] = -P[i]
    roots = polynomial.polyroots(coeffs)

    # Discard values where log is undefined
    roots = roots[np.logical_or(np.imag(roots) != 0.0, np.real(roots) >= 0.0)]

    # Deduce lamda and keep only interesting values
    lamda = np.real(np.log(roots)/(X[1] - X[0]))
    lamda = np.unique(lamda)
    lamda = lamda[np.logical_and(-20.0 < lamda, lamda < 0.0)]

    # Fit the values of 'a' on the curve
    def f(x, *ar):
        ar = np.asarray(ar)[:, np.newaxis]
        la = lamda[:, np.newaxis]
        return np.sum(ar * np.exp(la * x), axis=0)
    a, *_ = curve_fit(f, X, F, p0=np.zeros(lamda.shape))

    return a, lamda


def error_exponential_decomposition(X, F, a, lamda):
    """Mean square error of the exponential decomposition defined by the
    coefficients a and lamda with respect to the reference values in F.

    Parameters
    ----------
    X: 1D array
        sampling points
    F: 1D array (same size as X)
        reference values
    a: 1D array
        coefficients of the exponentials
    lamda: 1D array (same size as a)
        growth rate of the exponentials

    Returns
    -------
    error: float
        mean square error of the decomposition
    """
    a = np.asarray(a)[:, np.newaxis]
    lamda = np.asarray(lamda)[:, np.newaxis]

    def f(x):
        return np.sum(a * np.exp(lamda*x), axis=0)

    return np.square(f(X) - F).mean()


class PronyDecompositionFailure(Exception):
    pass


def find_best_exponential_decomposition(f, x_min, x_max, n_exp_range, *, tol=1e-4, noise_on_domain_points_std=0.01):
    """Tries to construct an exponential decompositoin of the function f on the
    domain [x_min, x_max] by testing the number of exponentials in n_exp_range.

    Parameters
    ----------
    f: callable
        The function ℝ→ℝ to be approximated.
        Should support vectorized calls (that is passing a vector of inputs
        and get the vector of corresponding outputs)
    x_min, x_max: floats
        The bounds of the domain of input in which f should be approximated
    n_exp_range: iterable of ints
        The decomposition sizes that will be tested
    tol: float, optional
        The target mean square error.
    noise_on_domain_points_std: float, optional
        Introduces some random variability on the points where the function is evaluated.
        Set this parameter to zero to disable randomness.

    """
    # Try different range of evaluation points to construct the decomposition.
    for n_exp in n_exp_range:

        # f might be ill-defined at some single specific values
        # (for the use-case of delhommeau.py, it is when x = kh exactly).
        # Thus we slightly randomize the range of evaluation points for the Prony decomposition.
        # This way, if one of the evaluation points hits the singular point, it will most likely not hit it again at the next iteration.
        x_max_iter = (1 + noise_on_domain_points_std*RNG.uniform())*x_max

        try:
            # The coefficients are computed on a resolution of 4*n_exp+1 ...
            X = np.linspace(x_min, x_max_iter, 4*n_exp+1)
            a, lamda = exponential_decomposition(X, f(X), n_exp)

            # ... and they are evaluated on a finer discretization.
            X = np.linspace(x_min, x_max_iter, 8*n_exp+1)
            if error_exponential_decomposition(X, f(X), a, lamda) < tol:
                return a, lamda
        except Exception:
            # If something bad happened while computing the decomposition, try
            # the next one.
            continue

    raise PronyDecompositionFailure(
            "No suitable Prony decomposition has been found in "
            f"[{x_min}, {x_max}] for tol={tol} "
            f"using a number of terms in {n_exp_range}."
            )



"""A simple timer class used to measure the time spent in various parts of the BEM solver."""

from functools import wraps
import time

class Timer:
    """A simple timer class that can be used as context manager or as decorator using `wraps_function` method

    Example
    -------
    ::

        timer = Timer()
        with timer:
            sleep(1.0)

        print(timer.total)  # 1.0...

        @timer.wraps_function
        def my_function():
            sleep(0.5)

        my_function()
        print(timer.total)  # 1.5...
        my_function()
        print(timer.total)  # 2.0...

        print(timer.timings)  # [1.0, 0.5, 0.5]
    """

    def __init__(self, timings=None):
        if timings is None:
            self.timings = []
        else:
            self.timings = timings

    def __repr__(self):
        return f"Timer({self.timings})"

    @property
    def nb_timings(self):
        return len(self.timings)

    @property
    def total(self):
        return sum(self.timings)

    @property
    def mean(self):
        if self.nb_timings == 0:
            return float('nan')
        else:
            return self.total/self.nb_timings

    def __enter__(self):
        self.start_time = time.perf_counter()

    def __exit__(self, *exc):
        self.timings.append(time.perf_counter() - self.start_time)

    def wraps_function(self, f):
        @wraps(f)
        def wrapped_f(*args, **kwargs):
            with self:
                return f(*args, **kwargs)
        return wrapped_f



"""Tests for the mesh submodule: definition and transformation of base meshes."""

import pytest

import numpy as np
from numpy.linalg import norm
from numpy.typing import NDArray

import capytaine as cpt
from capytaine.meshes.properties import clustering

RNG = np.random.default_rng()


# Some meshes that will be used in the following tests.
test_mesh = cpt.Mesh(vertices=RNG.random((4, 3)), faces=[range(4)], name="test_mesh")
cylinder = cpt.mesh_horizontal_cylinder()
sphere = cpt.mesh_sphere(radius=1)


def test_mesh_initialization():
    """Test how the code checks the validity of the parameters."""
    with pytest.raises(AssertionError):
        cpt.Mesh(vertices=RNG.random((6, 3)), faces=[(0, 1, 2), (3, 4, 5)])

    with pytest.raises(AssertionError):
        cpt.Mesh(vertices=RNG.random((4, 3)), faces=[(0, 1, 2, -1)])

    with pytest.raises(AssertionError):
        cpt.Mesh(vertices=RNG.random((3, 3)), faces=[(0, 1, 2, 3)])


def test_vertices_and_faces_get_and_set():
    """Test get and set functions for vertices and faces."""
    # Get faces
    faces = cylinder.faces
    vertices = cylinder.vertices

    new_cylinder = cpt.Mesh(name="new_cylinder")

    # Set faces
    with pytest.raises(AssertionError):
        new_cylinder.faces = faces
        # You should set the vertices first.
    new_cylinder.vertices = vertices
    new_cylinder.faces = faces

    assert new_cylinder == cylinder


def test_mesh_naming():
    """Test how the mesh handle names and string representation."""
    # Test automatic naming
    dummy_mesh = cpt.Mesh()  # Automatically named something like mesh_1
    other_dummy_mesh = cpt.Mesh()  # Automatically named something like mesh_2
    assert dummy_mesh.name[:5] == "mesh_"
    assert other_dummy_mesh.name[:5] == "mesh_"
    assert int(dummy_mesh.name[5:]) + 1 == int(other_dummy_mesh.name[5:])


def test_as_set_of_faces():
    """Test the representation of the mesh as a set of faces.
    Also allows to define the equality of two meshes."""
    faces = cylinder.as_set_of_faces()
    assert isinstance(faces, frozenset)
    assert len(faces) == cylinder.nb_faces
    assert all(len(face) == 4 or len(face) == 3 for face in faces)  # Each face is represented by 3 or 4 points
    assert all(len(vertex) == 3 for face in faces for vertex in face)  # Each point is represented by 3 coordinates.

    assert cylinder == cylinder  # The equality is defined as the equality of the set of faces.
    assert cpt.Mesh.from_set_of_faces(faces) == cylinder  # The mesh can be reconstructed from a set of faces.


def mesh_from_set_of_faces():
    A, B, C, D = (0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0)

    # Two triangular faces
    faces = {frozenset({A, B, C}), frozenset({B, C, D})}
    mesh = cpt.Mesh.from_set_of_faces(faces)
    assert mesh.nb_vertices == 4
    assert mesh.nb_faces == 2
    assert mesh.is_triangle(0) and mesh.is_triangle(1)


def test_copy():
    """Test the copy and renaming of meshes."""
    assert test_mesh.copy() is not test_mesh
    assert test_mesh.copy() == test_mesh
    assert test_mesh.copy(name="copy_of_the_test_mesh").name == "copy_of_the_test_mesh"
    assert test_mesh.merged() is test_mesh


def test_shape_helper_functions():
    assert np.allclose(sphere.center_of_mass_of_nodes, (0.0, 0.0, 0.0))
    assert np.isclose(sphere.diameter_of_nodes, 2.0)
    assert cylinder.squared_axis_aligned_bbox == (-5.0, 5.0, -5.0, 5.0, -5.0, 5.0)


def test_translate():
    assert cylinder.translated_x(0) == cylinder
    assert cylinder.translated_y(0) == cylinder
    assert cylinder.translated_z(0) == cylinder
    assert cylinder.translated([0, 0, 0]) == cylinder

    translated_sphere = sphere.translated([1, 2, 3])
    assert np.allclose(translated_sphere.center_of_mass_of_nodes, (1.0, 2.0, 3.0))
    assert np.isclose(translated_sphere.diameter_of_nodes, sphere.diameter_of_nodes)

def test_rotate():
    assert cylinder.rotated_x(0.0) == cylinder
    assert cylinder.rotated_y(0.0) == cylinder
    assert cylinder.rotated_z(0.0) == cylinder

    new_cylinder = cylinder.rotated_z(np.pi/2)
    assert np.allclose(new_cylinder.center_of_mass_of_nodes, cylinder.center_of_mass_of_nodes)
    assert new_cylinder.squared_axis_aligned_bbox == (-5.0, 5.0, -5.0, 5.0, -5.0, 5.0)


def test_mirror():
    new_cylinder = cylinder.mirrored(cpt.Plane())
    cylinder.mirror(cpt.Plane())
    assert new_cylinder == cylinder


def test_symmetrized():
    from capytaine.meshes.symmetric import ReflectionSymmetricMesh
    sym = cylinder.merged().symmetrized(cpt.xOz_Plane)
    assert isinstance(sym, ReflectionSymmetricMesh)


def test_mesh_quality():
    # TODO: Could be improved...
    cylinder.triangulate_quadrangles()
    cylinder.merge_duplicates(atol=1e-5)
    cylinder.heal_mesh()

def test_heal_mesh_removes_degenerate_panels():
    vertices = np.array([(0.0, 0.0, 0.0), (0.0, 1.0, 0.0), (1.0, 1.0, 0.0), (1.0, 0.0, 0.0)])
    faces = np.array([[0, 1, 2, 3], [1, 2, 2, 1]])
    mesh = cpt.Mesh(vertices, faces)
    assert mesh.nb_faces == 2
    mesh.heal_mesh()
    assert mesh.nb_faces == 1

@pytest.mark.xfail
def test_heal_mesh_with_complicated_connectivities():
    vertices = np.array([
        [-8.00000000e+00,  1.65358984e+00, -4.99999996e-02],
        [-8.00000000e+00,  1.65358984e+00,  5.00000003e-02],
        [-8.00000000e+00,  1.74019238e+00, -9.99999998e-02],
        [-8.00000000e+00,  1.74019238e+00, -1.78037182e-10],
        [-8.00000000e+00,  1.74019238e+00,  1.00000000e-01],
        [-8.00000000e+00,  1.82679492e+00, -5.00000002e-02],
        [-8.00000000e+00,  1.82679492e+00,  4.99999997e-02]
        ])
    faces = np.array([
        [5, 3, 6, 5],
        [5, 2, 3, 5],
        [3, 4, 6, 3],
        [3, 4, 6, 3],
        [2, 0, 3, 2],
        [0, 1, 3, 0],
        [1, 4, 3, 1]
        ])
    mesh = cpt.Mesh(vertices, faces)
    mesh.heal_mesh()

def test_extract_one_face():
    i = 2
    one_face = sphere.extract_one_face(i)
    assert np.all(one_face.faces_centers[0] == sphere.faces_centers[i])


def test_join_meshes_masks():
    a = cpt.mesh_sphere(resolution=(2, 3))
    b = cpt.mesh_sphere(center=(10, 0, 0), resolution=(3, 3))
    joined, masks = a.join_meshes(b, return_masks=True)
    assert np.all(masks[0] ^ masks[1])
    assert sum(masks[0]) == a.nb_faces
    assert sum(masks[1]) == b.nb_faces


##########################
#  Connected components  #
##########################

def test_connected_components_of_empty_mesh():
    from capytaine.meshes.properties import connected_components
    mesh = cpt.Mesh()
    cc = connected_components(mesh)
    assert len(cc) == 0


def test_connected_components_of_sphere():
    from capytaine.meshes.properties import connected_components
    mesh = cpt.mesh_sphere()
    cc = connected_components(mesh)
    assert len(cc) == 1


def test_connected_components_of_two_spheres():
    from capytaine.meshes.properties import connected_components
    mesh = cpt.mesh_sphere() + cpt.mesh_sphere(center=(5, 0, 0))
    cc = connected_components(mesh)
    assert len(cc) == 2


def test_connected_components_of_open_sphere():
    from capytaine.meshes.properties import connected_components
    mesh = cpt.mesh_sphere().immersed_part()
    cc = connected_components(mesh)
    assert len(cc) == 1


def test_connected_components_of_torus():
    from capytaine.io.meshio import load_from_meshio
    from capytaine.meshes.properties import connected_components
    pygmsh = pytest.importorskip("pygmsh")
    with pygmsh.occ.Geometry() as geom:
        geom.add_torus((0, 0, 0), 2.0, 0.5, mesh_size=0.3)
        gmsh_mesh = geom.generate_mesh(dim=2)
    mesh = load_from_meshio(gmsh_mesh).heal_mesh()
    cc = connected_components(mesh)
    assert len(cc) == 1


######################################
#  Connected components at waterline #
######################################

def test_connected_components_at_waterline_of_sphere():
    from capytaine.meshes.properties import connected_components_of_waterline
    mesh = cpt.mesh_sphere()
    cc = connected_components_of_waterline(mesh)
    assert len(cc) == 1


def test_connected_components_at_waterline_of_two_spheres():
    from capytaine.meshes.properties import connected_components_of_waterline
    mesh = cpt.mesh_sphere() + cpt.mesh_sphere(center=(5, 0, 0))
    cc = connected_components_of_waterline(mesh)
    assert len(cc) == 2


def test_connected_components_at_waterline_of_open_sphere():
    from capytaine.meshes.properties import connected_components_of_waterline
    mesh = cpt.mesh_sphere().immersed_part()
    cc = connected_components_of_waterline(mesh)
    assert len(cc) == 1


def test_connected_components_at_waterline_of_immersed_sphere():
    from capytaine.meshes.properties import connected_components_of_waterline
    mesh = cpt.mesh_sphere(center=(0, 0, -5))
    cc = connected_components_of_waterline(mesh, z=0.0)
    assert len(cc) == 0


def test_connected_components_at_waterline_of_torus():
    from capytaine.io.meshio import load_from_meshio
    from capytaine.meshes.properties import connected_components_of_waterline
    pygmsh = pytest.importorskip("pygmsh")
    with pygmsh.occ.Geometry() as geom:
        geom.add_torus((0, 0, 0), 2.0, 0.5, mesh_size=0.3)
        gmsh_mesh = geom.generate_mesh(dim=2)
    mesh = load_from_meshio(gmsh_mesh).heal_mesh()
    cc = connected_components_of_waterline(mesh)
    assert len(cc) == 2


###################################
#  Connected vertices clustering  #
###################################

@pytest.mark.parametrize("faces", [
    np.array([[1, 2], [1, 3], [2, 3], [4, 5], [5, 6], [7, 8], [3, 9]]),
    np.array([[0, 1, 2, 3], [1, 2, 6, 7]]),
])
def test_vertice_clustering(faces: NDArray[np.integer]):
    """Test the clustering algorithm for connected faces & vertices."""
    # Legacy way to cluster
    vertices_components: set[frozenset[int]] = set()
    for set_of_v_in_face in map(frozenset, faces):
        intersecting_components = [c for c in vertices_components if len(c.intersection(set_of_v_in_face)) > 0]
        if len(intersecting_components) == 0:
            vertices_components.add(set_of_v_in_face)
        else:
            for c in intersecting_components:
                vertices_components.remove(c)
            vertices_components.add(frozenset.union(set_of_v_in_face, *intersecting_components))
    # Vectorized clustering
    vert_groups = clustering(faces)
    assert {frozenset({*group}) for group in vert_groups} == vertices_components



import numpy as np
import xarray as xr
import capytaine as cpt
import matplotlib.pyplot as plt

cpt.set_logging('INFO')

# Initialize floating body by generating a geometric mesh
mesh = cpt.mesh_horizontal_cylinder(
    length=10.0, radius=0.5,  # Dimensions
    center=(0, 0, 0),         # Position
    resolution=(5, 20, 40)    # Fineness of the mesh
    ).immersed_part()
body = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs())

# Define the range of water depth
depth_range = np.concatenate([np.linspace(2, 100, 10), [np.inf]])

test_matrix = xr.Dataset(coords={
    "wavelength": 100.0,
    "water_depth": depth_range,
    "radiating_dof": ["Heave"]
    })

# Solve all radiation problems
solver = cpt.BEMSolver()
data = solver.fill_dataset(test_matrix, body)

# Note that the solver could not solve the most shallow case and returned NaN:
data.added_mass.sel(radiating_dof="Heave", influenced_dof="Heave", water_depth=2.0).values

# Plot the added mass of each dofs as a function of the water depth.
fig, ax = plt.subplots(layout="constrained")
ax.plot(
    data.water_depth * data.wavenumber,
    data['added_mass'].sel(radiating_dof="Heave", influenced_dof="Heave"),
    marker="s", label="Finite depth"
)
ax.hlines(
    data['added_mass'].sel(radiating_dof="Heave", influenced_dof="Heave", water_depth=np.inf),
    xmin=5, xmax=8, label="Infinite depth",
    linestyle="--"
)
ax.set(xlabel='wavenumber * water_depth', ylabel='Heave-heave added mass')
ax.legend()
plt.show()



import pytest

import numpy as np
import xarray as xr

import capytaine as cpt


def test_dof():
    nodes = np.array([[0, 0, 0], [0, 0, 1], [1, 0, 1], [1, 0, 0]])
    faces = np.array([[0, 1, 2, 3]])
    body = cpt.FloatingBody(cpt.Mesh(nodes, faces), name="one_face")
    assert body.dofs == {}

    body.add_translation_dof(direction=(1.0, 0.0, 0.0), name="1")
    assert np.allclose(body.dofs["1"], np.array([1.0, 0.0, 0.0]))

    body.add_translation_dof(direction=(0.0, 1.0, 0.0), name="2")
    assert np.allclose(body.dofs["2"], np.array([0.0, 1.0, 0.0]))

    body.add_rotation_dof(cpt.Axis(vector=(0.0, 0.0, 1.0)), name="3")
    body.add_rotation_dof(cpt.Axis(point=(0.5, 0, 0), vector=(0.0, 0.0, 1.0)), name="4")


def test_dof_name_inference():
    body = cpt.FloatingBody(mesh=cpt.mesh_horizontal_cylinder())
    body.add_translation_dof(direction=(1, 0, 0), name="Surge_1")
    for dofname in ['Surge', 'SURGE', 'surge']:
        body.add_translation_dof(name=dofname)
        assert np.allclose(body.dofs[dofname], body.dofs['Surge_1'])

    body.add_rotation_dof(name="Pitch")
    body.add_rotation_dof(name="yaw")

    body.dofs.clear()
    body.add_all_rigid_body_dofs()


def test_rigid_body_dofs():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(10.0, 0.0, 0.0)))
    assert "Heave" in body.dofs
    assert np.all(body.dofs["Pitch"][:, 2] > 9.0)


def test_rigid_body_dofs_no_rotation_center_but_a_center_of_mass():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(), center_of_mass=(-10.0, 0.0, 0.0))
    assert np.all(body.dofs["Pitch"][:, 2] < -9.0)


def test_rigid_body_dofs_both_a_rotation_center_and_a_center_of_mass():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(10.0, 0.0, 0.0)),
                            center_of_mass=(-10.0, 0.0, 0.0))
    assert np.all(body.dofs["Pitch"][:, 2] > 9.0)


def test_rigid_body_dofs_neither_a_rotation_center_nor_a_center_of_mass():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    assert np.allclose(body._infer_rotation_center(), (0.0, 0.0, 0.0))


def test_defining_rotation_center_with_ints():
    # Issue #319
    mesh = cpt.mesh_sphere().immersed_part()
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, -1)))
    body.translated_y(-2.0)


def test_healing_before_initializing_dofs():
    # Issue #367: https://github.com/capytaine/capytaine/issues/367
    # Define a mesh with a normal panel and degenerate panel
    vertices = np.array([(0.0, 0.0, 0.0), (0.0, 1.0, 0.0),
                         (1.0, 1.0, 0.0), (1.0, 0.0, 0.0)])
    faces = np.array([[0, 1, 2, 3], [1, 2, 2, 1]])
    mesh = cpt.Mesh(vertices, faces)
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    assert body.dofs["Heave"].shape[0] == body.mesh.nb_faces == 1


def test_bodies():
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere(), name="sphere", center_of_mass=(0, 0, 0))
    repr(body)
    body.add_translation_dof(name="Surge")
    body.add_translation_dof(name="Heave")

    # Extract faces
    body.extract_faces(np.where(body.mesh.faces_centers[:, 2] < 0)[0])

    # Clipping
    body.keep_immersed_part(inplace=False)

    # Mirror of the dofs
    mirrored = body.mirrored(cpt.Plane(point=(1, 0, 0), normal=(1, 0, 0)))
    assert np.allclose(mirrored.center_of_mass, np.array([2, 0, 0]))
    assert np.allclose(body.dofs['Surge'], -mirrored.dofs['Surge'])

    # Rotation of the dofs
    sideways = body.rotated(cpt.Axis(point=(0, 0, 0), vector=(0, 1, 0)), np.pi/2)
    assert np.allclose(sideways.dofs['Heave'][0], np.array([1, 0, 0]))

    upside_down = body.rotated(cpt.Axis(point=(0, 0, 0), vector=(0, 1, 0)), np.pi)
    assert np.allclose(body.dofs['Heave'], -upside_down.dofs['Heave'])

    # Copy of the body
    copy_of_body = body.copy(name="copy_of_sphere")
    copy_of_body.translate_x(10.0)
    copy_of_body.add_translation_dof(name="Heave")

    # Join bodies
    both = body.join_bodies(copy_of_body)
    assert set(both.dofs) == {'sphere__Surge', 'copy_of_sphere__Surge', 'sphere__Heave', 'copy_of_sphere__Heave'}


def test_mirror_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (0, 1, 0)
    body.mirror(cpt.xOz_Plane)
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([0, -1, 0]))


def test_translate_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (0, 0, 0)
    body.translate((0, 0, 1))
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([0, 0, 1]))


def test_rotate_rotation_center_defined_as_tuple():
    body = cpt.FloatingBody()
    body.rotation_center = (1, 0, 0)
    body.rotate_z(np.pi)
    assert body.rotation_center.shape == (3,)
    np.testing.assert_allclose(body.rotation_center, np.array([-1, 0, 0]), atol=1e-10)


def test_inplace_transform_with_lid():
    mesh = cpt.mesh_sphere().immersed_part()
    lid_mesh = mesh.generate_lid()
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh)
    body.mesh_including_lid
    body.translate_x(10.0)
    assert np.all(body.mesh.vertices[:, 0] > 5.0)
    assert np.all(body.mesh_including_lid.vertices[:, 0] > 5.0)


@pytest.mark.parametrize("z_center", [0, 2, -2])
@pytest.mark.parametrize("as_collection_of_meshes", [True, False])
def test_clipping_of_dofs(z_center, as_collection_of_meshes):
    """Check that clipping a body with a dof is the same as clipping the body ant then adding the dof."""
    full_sphere = cpt.FloatingBody(mesh=cpt.mesh_sphere(center=(0, 0, z_center), axial_symmetry=as_collection_of_meshes), name="sphere")
    axis = cpt.Axis(point=(1, 0, 0), vector=(1, 0, 0))

    full_sphere.add_rotation_dof(axis, name="test_dof")
    clipped_sphere = full_sphere.keep_immersed_part(free_surface=0.0, water_depth=np.inf, inplace=False)

    other_clipped_sphere = cpt.FloatingBody(mesh=clipped_sphere.mesh, name="other_sphere")
    other_clipped_sphere.add_rotation_dof(axis, name="test_dof")

    if clipped_sphere.mesh.nb_faces > 0:
        assert np.allclose(clipped_sphere.dofs['test_dof'], other_clipped_sphere.dofs['test_dof'])
    else:
        assert len(clipped_sphere.dofs['test_dof']) == 0


def test_complicated_clipping_of_dofs():
    # 1 face becomes 2 faces after clipping
    mesh = cpt.Mesh(vertices=[[0.0, 0.0, 0.5], [-0.5, 0.0, -0.5], [0.0, 0.0, -1.5], [0.0, 0.5, -0.5]], faces=[[0, 1, 2, 3]])
    body = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs())
    clipped_body = body.immersed_part()
    assert len(clipped_body.dofs["Heave"]) == clipped_body.mesh.nb_faces


def test_clipping_of_inconsistent_dof():
    body = cpt.FloatingBody(cpt.mesh_sphere(), dofs=cpt.rigid_body_dofs())
    body.mesh.keep_immersed_part()  # Dofs and mesh are now inconsistent
    with pytest.raises(ValueError):
        body.keep_immersed_part()


@pytest.mark.xfail
def test_clipping_of_dofs_with_degenerate_faces():
    vertices = np.array([
        [-8.00000000e+00,  1.65358984e+00, -4.99999996e-02],
        [-8.00000000e+00,  1.65358984e+00,  5.00000003e-02],
        [-8.00000000e+00,  1.74019238e+00, -9.99999998e-02],
        [-8.00000000e+00,  1.74019238e+00, -1.78037182e-10],
        [-8.00000000e+00,  1.74019238e+00,  1.00000000e-01],
        [-8.00000000e+00,  1.82679492e+00, -5.00000002e-02],
        [-8.00000000e+00,  1.82679492e+00,  4.99999997e-02]
        ])
    faces = np.array([
        [3, 4, 6, 3],
        [2, 0, 3, 2],
        ])
    mesh = cpt.Mesh(vertices, faces)
    body = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs())
    clipped_body = body.immersed_part()
    assert len(clipped_body.dofs["Heave"]) == clipped_body.mesh.nb_faces


def test_cropping_body_with_manual_dof():
    # https://github.com/capytaine/capytaine/issues/204
    sphere = cpt.FloatingBody(cpt.mesh_sphere())
    sphere.dofs["Surge"] = [(1, 0, 0) for face in sphere.mesh.faces]
    sphere.keep_immersed_part()


def test_immersed_part():
    full_sphere = cpt.FloatingBody(mesh=cpt.mesh_sphere(), name="ball")
    immersed_sphere = full_sphere.immersed_part()
    assert immersed_sphere is not full_sphere
    assert immersed_sphere.mesh == full_sphere.mesh.immersed_part()
    assert immersed_sphere.mesh.axis_aligned_bbox[5] <= 0.0
    assert immersed_sphere.name == "ball"
    full_sphere.translate_x(2.0)
    new_immersed_sphere = full_sphere.immersed_part()
    assert new_immersed_sphere is not immersed_sphere
    assert not np.allclose(new_immersed_sphere.mesh.axis_aligned_bbox, immersed_sphere.mesh.axis_aligned_bbox)


def test_mincing():
    body = cpt.FloatingBody(mesh=cpt.mesh_horizontal_cylinder(length=10, radius=0.5))
    body = body.minced((4, 1, 1))
    assert len(body.mesh) == 2
    assert np.all(body.mesh[0].faces_centers[:, 0] < 0)
    assert isinstance(body.mesh[0][0], cpt.Mesh)
    body = body.minced((1, 2, 2))
    assert isinstance(body.mesh[0][0][0][0], cpt.Mesh)


def test_assemble_regular_array():
    body = cpt.FloatingBody(mesh=cpt.mesh_sphere())
    body.add_all_rigid_body_dofs()
    array = body.assemble_regular_array(distance=2.0, nb_bodies=(2, 3))
    assert array.mesh.nb_faces == 6*body.mesh.nb_faces

    # Check name and order of the dofs
    assert list(array.dofs.keys())[0:3] == ["0_0__Surge", "0_0__Sway", "0_0__Heave"]
    assert "2_1__Heave" not in array.dofs.keys()

    # Check that the dofs corresponds to the right panels
    faces_1_0 = np.where(array.dofs["1_0__Heave"] != 0.0)[0]
    fc_1_0 = array.mesh.merged().faces_centers[faces_1_0, :]
    assert np.all(1.0 <= fc_1_0[:, 0]) and np.all(fc_1_0[:, 0] <= 3.0)  #   1 < x < 3
    assert np.all(-1.0 <= fc_1_0[:, 1]) and np.all(fc_1_0[:, 1] <= 1.0) #  -1 < y < 1


r = 1.0
locations = np.array([[1,0],[-1,0],[0,1]])*r*2
n_bodies = locations.shape[0]

@pytest.fixture
def fb_array():
    sphere = cpt.FloatingBody(cpt.mesh_sphere(radius=r, center=(0, 0, 0), resolution=(10, 4)))
    my_axis = cpt.Axis((0, 1, 0), point=(0,0,0))
    sphere.add_rotation_dof(axis=my_axis)
    sphere.keep_immersed_part()

    return sphere.assemble_arbitrary_array(locations)


def test_consistent_dofs_to_faces(fb_array):
    num_active_faces = []
    for fb_dof in fb_array.dofs.items():
        num_active_faces.append(np.count_nonzero(np.count_nonzero(fb_dof[1],axis=1)))

    ma = np.array(num_active_faces)

    tot_faces = fb_array.mesh.nb_faces
    exp_active_faces_per_dof = int(tot_faces/n_bodies)

    assert np.all(ma==exp_active_faces_per_dof)


def test_solve_hydrodynamics(fb_array):
    solver = cpt.BEMSolver()
    test_matrix = xr.Dataset(coords={
          'rho': 1e3,
          'water_depth': [np.inf],
          'omega': np.pi * 2 / 1,
          'wave_direction': 0,
          'radiating_dof': list(fb_array.dofs.keys()),
          })
    data = solver.fill_dataset(test_matrix, [fb_array],
                                 mesh=True,
                                 wavelength=True,
                                 wavenumber=True)
    assert data.influenced_dof.size == n_bodies
    assert data.radiating_dof.size == n_bodies
    assert data.added_mass.notnull().all()
    assert data.radiation_damping.notnull().all()
    assert data.diffraction_force.notnull().all()
    assert data.Froude_Krylov_force.notnull().all()


def test_cluster_bodies():
    centers = np.array([[0.0, 0.0, 0.0],
                        [10.0, 0.0, 0.0],
                        [11.0, 0.0, 0.0],
                        [10.0, 1.0, 0.0]])
    meshes = [cpt.mesh_sphere(center=c, name=str(c)) for c in centers]
    bodies = [cpt.FloatingBody(mesh=m) for m in meshes]
    joined_bodies = cpt.FloatingBody.join_bodies(*bodies)
    clustered_bodies = cpt.FloatingBody.cluster_bodies(*bodies)
    assert isinstance(clustered_bodies, cpt.FloatingBody)
    assert isinstance(clustered_bodies.mesh, cpt.CollectionOfMeshes)
    assert clustered_bodies.mesh.merged() == joined_bodies.mesh.merged()
    assert meshes[0] in clustered_bodies.mesh  # The first body is at the top level independently from the other three


def test_clip_component_of_multibody():
    # https://github.com/capytaine/capytaine/issues/660
    body_1 = cpt.FloatingBody(
        mesh=cpt.mesh_sphere(center=(0.0, 0.0, 0.0)),
        dofs=cpt.rigid_body_dofs(rotation_center=(0.0, 0.0, 0.0)),
        name="body_1"
    )
    body_2 = cpt.FloatingBody(
        mesh=cpt.mesh_sphere(center=(5.0, 0.0, 0.0)),
        dofs=cpt.rigid_body_dofs(rotation_center=(5.0, 0.0, 0.0)),
        name="body_2"
    )
    both = body_1 + body_2
    body_2.keep_immersed_part()
    assert both.dofs["body_1__Heave"].shape[0] == both.mesh.nb_faces



"""Tests for the mesh import of mesh from HAMS file format"""
from unittest import mock

import numpy as np

import capytaine as cpt
from capytaine.io.mesh_loaders import load_PNL
from capytaine.io.mesh_writers import write_PNL


def parse_pnl(file_content):
    # Workaround to avoid actually writing/reading a file on disk
    with mock.patch('capytaine.io.mesh_loaders._check_file', lambda foo: None):
        with mock.patch('capytaine.io.mesh_loaders.open', mock.mock_open(read_data=file_content)):
            return load_PNL("mocked/filename")


def test_pnl_no_symmetry():
    mesh = parse_pnl(
            """ --------------Hull Mesh File---------------

            # Number of Panels, Nodes, X-Symmetry and Y-Symmetry
            2        8           0           0

            #Start Definition of Node Coordinates     ! node_number   x   y   z
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            #End Definition of Node Coordinates

            #Start Definition of Node Relations
            1 4  1 2 3 4
            2 4  5 6 7 8
            #End Definition of Node Relations
            """)
    assert isinstance(mesh, cpt.Mesh)
    assert mesh.nb_faces == 2


def test_pnl_triangle():
    mesh = parse_pnl(
            """ --------------Hull Mesh File---------------

            # Number of Panels, Nodes, X-Symmetry and Y-Symmetry
            2        7           0           0

            #Start Definition of Node Coordinates     ! node_number   x   y   z
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            #End Definition of Node Coordinates

            #Start Definition of Node Relations
            1 4  1 2 3 4
            2 3  5 6 7
            #End Definition of Node Relations
            """)
    assert isinstance(mesh, cpt.Mesh)
    assert mesh.nb_faces == 2


def test_pnl_x_symmetry():
    mesh = parse_pnl(
            """ --------------Hull Mesh File---------------

            # Number of Panels, Nodes, X-Symmetry and Y-Symmetry
            2        8           1           0

            #Start Definition of Node Coordinates     ! node_number   x   y   z
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            #End Definition of Node Coordinates

            #Start Definition of Node Relations
            1 4  1 2 3 4
            2 4  5 6 7 8
            #End Definition of Node Relations
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert mesh.plane == cpt.yOz_Plane
    assert mesh.nb_faces == 4


def test_pnl_y_symmetry():
    mesh = parse_pnl(
            """ --------------Hull Mesh File---------------

            # Number of Panels, Nodes, X-Symmetry and Y-Symmetry
            2        8           0           1

            #Start Definition of Node Coordinates     ! node_number   x   y   z
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            #End Definition of Node Coordinates

            #Start Definition of Node Relations
            1 4  1 2 3 4
            2 4  5 6 7 8
            #End Definition of Node Relations
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert mesh.plane == cpt.xOz_Plane
    assert mesh.nb_faces == 4


def test_pnl_xy_symmetry():
    mesh = parse_pnl(
            """ --------------Hull Mesh File---------------

            # Number of Panels, Nodes, X-Symmetry and Y-Symmetry
            2        8           1           1

            #Start Definition of Node Coordinates     ! node_number   x   y   z
            1 0.0 0.0 -1.0
            2 1.0 0.0 -1.0
            3 1.0 1.0 -1.0
            4 0.0 1.0 -1.0
            5 1.0 0.0 -1.0
            6 2.0 0.0 -1.0
            7 2.0 1.0 -1.0
            8 1.0 1.0 -1.0
            #End Definition of Node Coordinates

            #Start Definition of Node Relations
            1 4  1 2 3 4
            2 4  5 6 7 8
            #End Definition of Node Relations
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.ReflectionSymmetricMesh)
    assert mesh.nb_faces == 8


###################################################


def test_pnl_roundtrip(tmpdir):
    mesh_path = tmpdir.join("temp_mesh.pnl")
    mesh = cpt.mesh_sphere()
    write_PNL(mesh_path, mesh.vertices, mesh.faces)
    reloaded_mesh = cpt.load_mesh(mesh_path)
    np.testing.assert_equal(mesh.faces, reloaded_mesh.faces)
    np.testing.assert_allclose(mesh.vertices, reloaded_mesh.vertices, atol=1e-5)



"""Floating bodies to be used in radiation-diffraction problems."""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import logging
import copy
from itertools import chain, accumulate, zip_longest
from functools import cached_property, lru_cache

import numpy as np
import xarray as xr

from capytaine.meshes.mesh_like_protocol import MeshLike
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.geometry import Abstract3DObject, ClippableMixin, Plane, inplace_transformation
from capytaine.meshes.properties import connected_components, connected_components_of_waterline
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.symmetric import build_regular_array_of_meshes
from capytaine.bodies.dofs import RigidBodyDofsPlaceholder

from capytaine.tools.optional_imports import silently_import_optional_dependency
meshio = silently_import_optional_dependency("meshio")

LOG = logging.getLogger(__name__)

TRANSLATION_DOFS_DIRECTIONS = {"surge": (1, 0, 0), "sway": (0, 1, 0), "heave": (0, 0, 1)}
ROTATION_DOFS_AXIS = {"roll": (1, 0, 0), "pitch": (0, 1, 0), "yaw": (0, 0, 1)}


class FloatingBody(ClippableMixin, Abstract3DObject):
    """A floating body described as a mesh and some degrees of freedom.

    The mesh structure is stored as a Mesh from capytaine.mesh.mesh or a
    CollectionOfMeshes from capytaine.mesh.meshes_collection.

    The degrees of freedom (dofs) are stored as a dict associating a name to
    a complex-valued array of shape (nb_faces, 3). To each face of the body
    (as indexed in the mesh) corresponds a complex-valued 3d vector, which
    defines the displacement of the center of the face in frequency domain.

    Parameters
    ----------
    mesh : MeshLike, optional
        the mesh describing the geometry of the hull of the floating body.
        If none is given, a empty one is created.
    lid_mesh : MeshLike or None, optional
        a mesh of an internal lid for irregular frequencies removal.
        Unlike the mesh of the hull, no dof is defined on the lid_mesh.
        If none is given, none is used when solving the Boundary Integral Equation.
    dofs : dict, optional
        the degrees of freedom of the body.
        If none is given, a empty dictionary is initialized.
    mass : float or None, optional
        the mass of the body in kilograms.
        Required only for some hydrostatics computation.
        If None, the mass is implicitly assumed to be the mass of displaced water.
    center_of_mass: 3-element array, optional
        the position of the center of mass.
        Required only for some hydrostatics computation.
    name : str, optional
        a name for the body.
        If none is given, the one of the mesh is used.
    """

    def __init__(self, mesh=None, dofs=None, mass=None, center_of_mass=None, name=None, *, lid_mesh=None):
        if mesh is None:
            self.mesh = Mesh(name="dummy_mesh")

        elif meshio is not None and isinstance(mesh, meshio._mesh.Mesh):
            from capytaine.io.meshio import load_from_meshio
            self.mesh = load_from_meshio(mesh)

        elif isinstance(mesh, MeshLike):
            self.mesh = mesh

        else:
            raise TypeError("Unrecognized `mesh` object passed to the FloatingBody constructor.")

        if lid_mesh is not None:
            if lid_mesh.nb_faces == 0:
                LOG.warning("Lid mesh %s provided for body initialization is empty. The lid mesh is ignored.", lid_mesh)
                self.lid_mesh = None
            else:
                self.lid_mesh = lid_mesh.with_normal_vector_going_down(inplace=False)
        else:
            self.lid_mesh = None

        if name is None and mesh is None:
            self.name = "dummy_body"
        elif name is None:
            if hasattr(self.mesh, "name"):
                self.name = self.mesh.name
            else:
                self.name = "anonymous_body"
        else:
            self.name = name

        self.mass = mass
        if center_of_mass is not None:
            self.center_of_mass = np.asarray(center_of_mass, dtype=float)
        else:
            self.center_of_mass = None

        if hasattr(self.mesh, "heal_mesh") and self.mesh.nb_vertices > 0 and self.mesh.nb_faces > 0:
            self.mesh.heal_mesh()

        if dofs is None:
            self.dofs = {}
        elif isinstance(dofs, RigidBodyDofsPlaceholder):
            if dofs.rotation_center is not None:
                self.rotation_center = np.asarray(dofs.rotation_center, dtype=float)
            self.dofs = {}
            self.add_all_rigid_body_dofs()
        else:
            self.dofs = dofs

        self._evaluate_full_mesh()

        LOG.info(f"New floating body: {self.__str__()}.")

        self._check_dofs_shape_consistency()

    def _evaluate_full_mesh(self):
        """Merge the mesh and lid_mesh, while keeping track of where each panel came from."""
        if self.lid_mesh is None:
            self.mesh_including_lid = self.mesh
            self.hull_mask = np.full((self.mesh.nb_faces,), True)
        else:
            self.mesh_including_lid, masks = self.mesh.join_meshes(self.lid_mesh, return_masks=True)
            self.hull_mask = masks[0]

    @staticmethod
    def from_meshio(mesh, name=None) -> 'FloatingBody':
        """Create a FloatingBody from a meshio mesh object.
        Kinda deprecated, use cpt.load_mesh instead."""
        from capytaine.io.meshio import load_from_meshio
        return FloatingBody(mesh=load_from_meshio(mesh, name), name=name)

    @staticmethod
    def from_file(filename: str, file_format=None, name=None) -> 'FloatingBody':
        """Create a FloatingBody from a mesh file using meshmagick.
        Kinda deprecated, use cpt.load_mesh instead."""
        from capytaine.io.mesh_loaders import load_mesh
        if name is None: name = filename
        mesh = load_mesh(filename, file_format, name=f"{name}_mesh")
        return FloatingBody(mesh, name=name)

    def __lt__(self, other: 'FloatingBody') -> bool:
        """Arbitrary order. The point is to sort together the problems involving the same body."""
        return self.name < other.name

    ##########
    #  Dofs  #
    ##########

    @property
    def nb_dofs(self) -> int:
        """Number of degrees of freedom."""
        return len(self.dofs)

    def add_translation_dof(self, direction=None, name=None, amplitude=1.0) -> None:
        """Add a new translation dof (in place).
        If no direction is given, the code tries to infer it from the name.

        Parameters
        ----------
        direction : array of shape (3,), optional
            the direction of the translation
        name : str, optional
            a name for the degree of freedom
        amplitude : float, optional
            amplitude of the dof (default: 1.0 m/s)
        """
        if direction is None:
            if name is not None and name.lower() in TRANSLATION_DOFS_DIRECTIONS:
                direction = TRANSLATION_DOFS_DIRECTIONS[name.lower()]
            else:
                raise ValueError("A direction needs to be specified for the dof.")

        if name is None:
            name = f"dof_{self.nb_dofs}_translation"

        direction = np.asarray(direction)
        assert direction.shape == (3,)

        motion = np.empty((self.mesh.nb_faces, 3))
        motion[:, :] = direction
        self.dofs[name] = amplitude * motion

    def add_rotation_dof(self, axis=None, name=None, amplitude=1.0) -> None:
        """Add a new rotation dof (in place).
        If no axis is given, the code tries to infer it from the name.

        Parameters
        ----------
        axis: Axis, optional
            the axis of the rotation
        name : str, optional
            a name for the degree of freedom
        amplitude : float, optional
            amplitude of the dof (default: 1.0)
        """
        if axis is None:
            if name is not None and name.lower() in ROTATION_DOFS_AXIS:
                axis_direction = ROTATION_DOFS_AXIS[name.lower()]
                for point_attr in ('rotation_center', 'center_of_mass', 'geometric_center'):
                    if hasattr(self, point_attr) and getattr(self, point_attr) is not None:
                        axis_point = getattr(self, point_attr)
                        LOG.info(f"The rotation dof {name} has been initialized around the point: "
                                 f"{self.__short_str__()}.{point_attr} = {getattr(self, point_attr)}")
                        break
                else:
                    axis_point = np.array([0, 0, 0])
                    LOG.warning(f"The rotation dof {name} has been initialized "
                                f"around the origin of the domain (0, 0, 0).")
            else:
                raise ValueError("A direction needs to be specified for the dof.")
        else:
            axis_point = axis.point
            axis_direction = axis.vector

        if name is None:
            name = f"dof_{self.nb_dofs}_rotation"

        if self.mesh.nb_faces == 0:
            self.dofs[name] = np.empty((self.mesh.nb_faces, 3))
        else:
            motion = np.cross(axis_point - self.mesh.faces_centers, axis_direction)
            self.dofs[name] = amplitude * motion

    def add_all_rigid_body_dofs(self) -> None:
        """Add the six degrees of freedom of rigid bodies (in place)."""
        self.add_translation_dof(name="Surge")
        self.add_translation_dof(name="Sway")
        self.add_translation_dof(name="Heave")
        self.add_rotation_dof(name="Roll")
        self.add_rotation_dof(name="Pitch")
        self.add_rotation_dof(name="Yaw")

    def integrate_pressure(self, pressure):
        forces = {}
        for dof_name in self.dofs:
            # Scalar product on each face:
            normal_dof_amplitude_on_face = - np.sum(self.dofs[dof_name] * self.mesh.faces_normals, axis=1)
            # The minus sign in the above line is because we want the force of the fluid on the body and not the force of the body on the fluid.
            # Sum over all faces:
            forces[dof_name] = np.sum(pressure * normal_dof_amplitude_on_face * self.mesh.faces_areas)
        return forces

    @inplace_transformation
    def keep_only_dofs(self, dofs):
        for dof in list(self.dofs.keys()):
            if dof not in dofs:
                del self.dofs[dof]

        if hasattr(self, 'inertia_matrix'):
            self.inertia_matrix = self.inertia_matrix.sel(radiating_dof=dofs, influenced_dof=dofs)
        if hasattr(self, 'hydrostatic_stiffness'):
            self.hydrostatic_stiffness = self.hydrostatic_stiffness.sel(radiating_dof=dofs, influenced_dof=dofs)

        return self

    def add_dofs_labels_to_vector(self, vector):
        """Helper function turning a bare vector into a vector labelled by the name of the dofs of the body,
        to be used for instance for the computation of RAO."""
        return xr.DataArray(data=np.asarray(vector), dims=['influenced_dof'],
                            coords={'influenced_dof': list(self.dofs)},
                            )

    def add_dofs_labels_to_matrix(self, matrix):
        """Helper function turning a bare matrix into a matrix labelled by the name of the dofs of the body,
        to be used for instance for the computation of RAO."""
        return xr.DataArray(data=np.asarray(matrix), dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': list(self.dofs), 'radiating_dof': list(self.dofs)},
                            )

    def _check_dofs_shape_consistency(self):
        for dof_name, dof in self.dofs.items():
            if np.array(dof).shape != (self.mesh.nb_faces, 3):
                raise ValueError(f"The array defining the dof {dof_name} of body {self.name} does not have the expected shape.\n"
                                 f"Expected shape: ({self.mesh.nb_faces}, 3)\n"
                                 f"  Actual shape: {dof.shape}")


    ###################
    # Hydrostatics #
    ###################

    def surface_integral(self, data, **kwargs):
        """Returns integral of given data along wet surface area."""
        return self.mesh.surface_integral(data, **kwargs)

    def waterplane_integral(self, data, **kwargs):
        """Returns integral of given data along water plane area."""
        return self.mesh.waterplane_integral(data, **kwargs)

    @property
    def wet_surface_area(self):
        """Returns wet surface area."""
        return self.mesh.wet_surface_area

    @property
    def volumes(self):
        """Returns volumes using x, y, z components of the FloatingBody."""
        return self.mesh.volumes

    @property
    def volume(self):
        """Returns volume of the FloatingBody."""
        return self.mesh.volume

    def disp_mass(self, *, rho=1000.0):
        return self.mesh.disp_mass(rho=rho)

    @property
    def center_of_buoyancy(self):
        """Returns center of buoyancy of the FloatingBody."""
        return self.mesh.center_of_buoyancy

    @property
    def waterplane_area(self):
        """Returns water plane area of the FloatingBody."""
        return self.mesh.waterplane_area

    @property
    def waterplane_center(self):
        """Returns water plane center of the FloatingBody.

        Note: Returns None if the FloatingBody is full submerged.
        """
        return self.mesh.waterplane_center

    @property
    def transversal_metacentric_radius(self):
        """Returns transversal metacentric radius of the mesh."""
        inertia_moment = -self.waterplane_integral(self.mesh.faces_centers[:,1]**2)
        return inertia_moment / self.volume

    @property
    def longitudinal_metacentric_radius(self):
        """Returns longitudinal metacentric radius of the mesh."""
        inertia_moment = -self.waterplane_integral(self.mesh.faces_centers[:,0]**2)
        return inertia_moment / self.volume

    @property
    def transversal_metacentric_height(self):
        """Returns transversal metacentric height of the mesh."""
        gb = self.center_of_mass - self.center_of_buoyancy
        return self.transversal_metacentric_radius - gb[2]

    @property
    def longitudinal_metacentric_height(self):
        """Returns longitudinal metacentric height of the mesh."""
        gb = self.center_of_mass - self.center_of_buoyancy
        return self.longitudinal_metacentric_radius - gb[2]

    def dof_normals(self, dof):
        """Returns dot product of the surface face normals and DOF"""
        return np.sum(self.mesh.faces_normals * dof, axis=1)

    def _infer_rotation_center(self):
        """Hacky way to infer the point around which the rotation dofs are defined.
        (Assuming all three rotation dofs are defined around the same point).
        In the future, should be replaced by something more robust.
        """
        if hasattr(self, "rotation_center"):
            return np.asarray(self.rotation_center)

        else:
            try:
                xc1 = self.dofs["Pitch"][:, 2] + self.mesh.faces_centers[:, 0]
                xc2 = -self.dofs["Yaw"][:, 1] + self.mesh.faces_centers[:, 0]
                yc1 = self.dofs["Yaw"][:, 0] + self.mesh.faces_centers[:, 1]
                yc2 = -self.dofs["Roll"][:, 2] + self.mesh.faces_centers[:, 1]
                zc1 = -self.dofs["Pitch"][:, 0] + self.mesh.faces_centers[:, 2]
                zc2 = self.dofs["Roll"][:, 1] + self.mesh.faces_centers[:, 2]

                # All items should be identical in a given vector
                assert np.isclose(xc1, xc1[0]).all()
                assert np.isclose(yc1, yc1[0]).all()
                assert np.isclose(zc1, zc1[0]).all()

                # Both vector should be identical
                assert np.allclose(xc1, xc2)
                assert np.allclose(yc1, yc2)
                assert np.allclose(zc1, zc2)

                return np.array([xc1[0], yc1[0], zc1[0]])

            except Exception as e:
                raise ValueError(
                        f"Failed to infer the rotation center of {self.name} to compute rigid body hydrostatics.\n"
                        f"Possible fix: add a `rotation_center` attribute to {self.name}.\n"
                        "Note that rigid body hydrostatic methods currently assume that the three rotation dofs have the same rotation center."
                        ) from e

    def each_hydrostatic_stiffness(self, influenced_dof_name, radiating_dof_name, *,
                                         influenced_dof_div=0.0, rho=1000.0, g=9.81):
        r"""
        Return the hydrostatic stiffness for a pair of DOFs.

        :math:`C_{ij} = \rho g\iint_S (\hat{n} \cdot V_j) (w_i + z D_i) dS`

        where :math:`\hat{n}` is surface normal,

        :math:`V_i = u_i \hat{n}_x + v_i \hat{n}_y + w_i \hat{n}_z` is DOF vector and

        :math:`D_i = \nabla \cdot V_i` is the divergence of the DOF.

        Parameters
        ----------
        influenced_dof_name : str
            Name of influenced DOF vector of the FloatingBody
        radiating_dof_name: str
            Name of radiating DOF vector of the FloatingBody
        influenced_dof_div: np.ndarray (Face_count), optional
            Influenced DOF divergence of the FloatingBody, by default 0.0.
        rho: float, optional
            water density, by default 1000.0
        g: float, optional
            Gravity acceleration, by default 9.81

        Returns
        -------
        hs_ij: xarray.variable
            hydrostatic_stiffness of ith DOF and jth DOF.

        Note
        ----
            This function computes the hydrostatic stiffness assuming :math:`D_{i} = 0`.
            If :math:`D_i \neq 0`, input the divergence interpolated to face centers.

            General integral equations are used for the rigid body modes and
            Neumann (1994) method is used for flexible modes.

        References
        ----------
            Newman, John Nicholas. "Wave effects on deformable bodies."Applied ocean
            research" 16.1 (1994): 47-59.
            http://resolver.tudelft.nl/uuid:0adff84c-43c7-43aa-8cd8-d4c44240bed8

        """
        # Newman (1994) formula is not 'complete' as recovering the rigid body
        # terms is not possible. https://doi.org/10.1115/1.3058702.

        # Alternative is to use the general equation of hydrostatic and
        # restoring coefficient for rigid modes and use Newman equation for elastic
        # modes.

        rigid_dof_names = ("Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw")
        dof_pair = (influenced_dof_name, radiating_dof_name)

        if set(dof_pair).issubset(set(rigid_dof_names)):
            if self.center_of_mass is None:
                raise ValueError(f"Trying to compute rigid-body hydrostatic stiffness for {self.name}, but no center of mass has been defined.\n"
                                 f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")
            mass = self.disp_mass(rho=rho) if self.mass is None else self.mass
            xc, yc, zc = self._infer_rotation_center()

            if dof_pair == ("Heave", "Heave"):
                norm_hs_stiff = self.waterplane_area
            elif dof_pair in [("Heave", "Roll"), ("Roll", "Heave")]:
                norm_hs_stiff = -self.waterplane_integral(self.mesh.faces_centers[:,1] - yc)
            elif dof_pair in [("Heave", "Pitch"), ("Pitch", "Heave")]:
                norm_hs_stiff = self.waterplane_integral(self.mesh.faces_centers[:,0] - xc)
            elif dof_pair == ("Roll", "Roll"):
                norm_hs_stiff = (
                        -self.waterplane_integral((self.mesh.faces_centers[:,1] - yc)**2)
                        + self.volume*(self.center_of_buoyancy[2] - zc) - mass/rho*(self.center_of_mass[2] - zc)
                )
            elif dof_pair in [("Roll", "Pitch"), ("Pitch", "Roll")]:
                norm_hs_stiff = self.waterplane_integral((self.mesh.faces_centers[:,0] - xc)
                                                          * (self.mesh.faces_centers[:,1] - yc))
            elif dof_pair == ("Roll", "Yaw"):
                norm_hs_stiff = - self.volume*(self.center_of_buoyancy[0] - xc) + mass/rho*(self.center_of_mass[0] - xc)
            elif dof_pair == ("Pitch", "Pitch"):
                norm_hs_stiff = (
                        -self.waterplane_integral((self.mesh.faces_centers[:,0] - xc)**2)
                        + self.volume*(self.center_of_buoyancy[2] - zc) - mass/rho*(self.center_of_mass[2] - zc)
                        )
            elif dof_pair == ("Pitch", "Yaw"):
                norm_hs_stiff = - self.volume*(self.center_of_buoyancy[1] - yc) + mass/rho*(self.center_of_mass[1] - yc)
            else:
                norm_hs_stiff = 0.0
        else:
            if self.mass is not None and not np.isclose(self.mass, self.disp_mass(rho=rho), rtol=1e-4):
                raise NotImplementedError(
                        f"Trying to compute the hydrostatic stiffness for dofs {radiating_dof_name} and {influenced_dof_name}"
                        f"of body {self.name}, which is not neutrally buoyant (mass={self.mass}, disp_mass={self.disp_mass(rho=rho)}).\n"
                        f"This case has not been implemented in Capytaine. You need either a single rigid body or a neutrally buoyant body."
                        )

            # Newman (1994) formula for flexible DOFs
            influenced_dof = np.array(self.dofs[influenced_dof_name])
            radiating_dof = np.array(self.dofs[radiating_dof_name])
            influenced_dof_div_array = np.array(influenced_dof_div)

            radiating_dof_normal = self.dof_normals(radiating_dof)
            z_influenced_dof_div = influenced_dof[:,2] + self.mesh.faces_centers[:,2] * influenced_dof_div_array
            norm_hs_stiff = self.surface_integral( -radiating_dof_normal * z_influenced_dof_div)

        hs_stiff = rho * g * norm_hs_stiff

        return xr.DataArray([[hs_stiff]],
                            dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': [influenced_dof_name],
                            'radiating_dof': [radiating_dof_name]},
                            name="hydrostatic_stiffness"
                            )

    def compute_hydrostatic_stiffness(self, *, divergence=None, rho=1000.0, g=9.81):
        r"""
        Compute hydrostatic stiffness matrix for all DOFs of the body.

        :math:`C_{ij} = \rho g\iint_S (\hat{n} \cdot V_j) (w_i + z D_i) dS`

        where :math:`\hat{n}` is surface normal,

        :math:`V_i = u_i \hat{n}_x + v_i \hat{n}_y + w_i \hat{n}_z` is DOF vector and

        :math:`D_i = \nabla \cdot V_i` is the divergence of the DOF.

        Parameters
        ----------
        divergence : dict mapping a dof name to an array of shape (nb_faces) or
                        xarray.DataArray of shape (nb_dofs × nb_faces), optional
            Divergence of the DOFs, by default None
        rho : float, optional
            Water density, by default 1000.0
        g: float, optional
            Gravity acceleration, by default 9.81

        Returns
        -------
        xr.DataArray
            Matrix of hydrostatic stiffness

        Note
        ----
            This function computes the hydrostatic stiffness assuming :math:`D_{i} = 0`.
            If :math:`D_i \neq 0`, input the divergence interpolated to face centers.

            General integral equations are used for the rigid body modes and
            Neumann (1994) method is used for flexible modes.

        References
        ----------
            Newman, John Nicholas. "Wave effects on deformable bodies."Applied ocean
            research" 16.1 (1994): 47-59.
            http://resolver.tudelft.nl/uuid:0adff84c-43c7-43aa-8cd8-d4c44240bed8

        """
        if len(self.dofs) == 0:
            raise AttributeError("Cannot compute hydrostatics stiffness on {} since no dof has been defined.".format(self.name))

        def divergence_dof(influenced_dof):
            if influenced_dof.lower() in [*TRANSLATION_DOFS_DIRECTIONS, *ROTATION_DOFS_AXIS]:
                return 0.0  # Dummy value that is not actually used afterwards.
            elif divergence is None:
                return 0.0
            elif isinstance(divergence, dict) and influenced_dof in divergence.keys():
                return divergence[influenced_dof]
            elif isinstance(divergence, xr.DataArray) and influenced_dof in divergence.coords["influenced_dof"]:
                return divergence.sel(influenced_dof=influenced_dof).values
            else:
                LOG.warning("Computing hydrostatic stiffness without the divergence of {}".format(influenced_dof))
                return 0.0

        hs_set =  xr.merge([
            self.each_hydrostatic_stiffness(
                influenced_dof_name, radiating_dof_name,
                influenced_dof_div = divergence_dof(influenced_dof_name),
                rho=rho, g=g
                )
            for radiating_dof_name in self.dofs
            for influenced_dof_name in self.dofs
            ], compat='no_conflicts', join="outer")

        # Reorder dofs
        K = hs_set.hydrostatic_stiffness.sel(influenced_dof=list(self.dofs.keys()), radiating_dof=list(self.dofs.keys()))
        return K

    def compute_rigid_body_inertia(self, *, rho=1000.0, output_type="body_dofs"):
        """
        Inertia Mass matrix of the body for 6 rigid DOFs.

        Parameters
        ----------
        rho : float, optional
            Density of water, by default 1000.0
        output_type : {"body_dofs", "rigid_dofs", "all_dofs"}
            Type of DOFs for mass mat output, by default "body_dofs".

        Returns
        -------
        xarray.DataArray
            Inertia matrix

        Raises
        ------
        ValueError
            If output_type is not in {"body_dofs", "rigid_dofs", "all_dofs"}.
        """
        if self.center_of_mass is None:
            raise ValueError(f"Trying to compute rigid-body inertia matrix for {self.name}, but no center of mass has been defined.\n"
                             f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")

        rc = self._infer_rotation_center()
        fcs = (self.mesh.faces_centers - rc).T
        combinations = np.array([fcs[0]**2, fcs[1]**2, fcs[2]**2, fcs[0]*fcs[1],
                                 fcs[1]*fcs[2], fcs[2]*fcs[0]])
        integrals = np.array([
            [np.sum(normal_i * fcs[axis] * combination * self.mesh.faces_areas)
            for combination in combinations]
            for axis, normal_i in enumerate(self.mesh.faces_normals.T)])


        inertias = np.array([
            (integrals[0,1]   + integrals[0,2]   + integrals[1,1]/3
             + integrals[1,2]   + integrals[2,1] + integrals[2,2]/3)/3,
            (integrals[0,0]/3 + integrals[0,2]   + integrals[1,0]
             + integrals[1,2]   + integrals[2,0] + integrals[2,2]/3)/3,
            (integrals[0,0]/3 + integrals[0,1]   + integrals[1,0]
             + integrals[1,1]/3 + integrals[2,0] + integrals[2,1]  )/3,
            integrals[2,3],
            integrals[0,4],
            integrals[1,5]
        ])

        cog = self.center_of_mass - rc
        volume = self.volume
        volumic_inertia_matrix = np.array([
            [ volume        , 0              , 0               ,
              0             , volume*cog[2]  , -volume*cog[1]  ],
            [ 0             , volume         , 0               ,
             -volume*cog[2] , 0              , volume*cog[0]   ],
            [ 0             , 0              , volume          ,
              volume*cog[1] , -volume*cog[0] , 0 ]             ,
            [ 0             , -volume*cog[2] , volume*cog[1]   ,
              inertias[0]   , -inertias[3]   , -inertias[5]    ],
            [ volume*cog[2] , 0              , -volume*cog[0]  ,
             -inertias[3]   , inertias[1]    , -inertias[4]    ],
            [-volume*cog[1] , volume*cog[0]  , 0               ,
             -inertias[5]   , -inertias[4]   , inertias[2]     ],
        ])

        density = rho if self.mass is None else self.mass/volume
        inertia_matrix = density * volumic_inertia_matrix

        # Rigid DOFs
        rigid_dof_names = ["Surge", "Sway", "Heave", "Roll", "Pitch", "Yaw"]
        rigid_inertia_matrix_xr = xr.DataArray(data=np.asarray(inertia_matrix),
                            dims=['influenced_dof', 'radiating_dof'],
                            coords={'influenced_dof': rigid_dof_names,
                                    'radiating_dof': rigid_dof_names},
                            name="inertia_matrix")

        # Body DOFs (Default as np.nan)
        body_dof_names = list(self.dofs)
        body_dof_count = len(body_dof_names)
        other_dofs_inertia_matrix_xr = xr.DataArray(np.nan * np.zeros([body_dof_count, body_dof_count]),
                                    dims=['influenced_dof', 'radiating_dof'],
                                    coords={'influenced_dof': body_dof_names,
                                            'radiating_dof': body_dof_names},
                                    name="inertia_matrix")

        total_mass_xr = xr.merge([rigid_inertia_matrix_xr, other_dofs_inertia_matrix_xr], compat="override", join="outer").inertia_matrix

        non_rigid_dofs = set(body_dof_names) - set(rigid_dof_names)

        if output_type == "body_dofs":
            if len(non_rigid_dofs) > 0:
                LOG.warning(f"Non-rigid dofs {non_rigid_dofs} detected: their \
inertia coefficients are assigned as NaN.")

            inertia_matrix_xr = total_mass_xr.sel(influenced_dof=body_dof_names,
                                                  radiating_dof=body_dof_names)
        elif output_type == "rigid_dofs":
            inertia_matrix_xr = total_mass_xr.sel(influenced_dof=rigid_dof_names,
                                                  radiating_dof=rigid_dof_names)
        elif output_type == "all_dofs":
            if len(non_rigid_dofs) > 0:
                LOG.warning("Non-rigid dofs: {non_rigid_dofs} are detected and \
respective inertia coefficients are assigned as NaN.")

            inertia_matrix_xr = total_mass_xr
        else:
            raise ValueError(f"output_type should be either 'body_dofs', \
'all_dofs' or 'rigid_dofs'. Given output_type = '{output_type}'.")

        return inertia_matrix_xr


    def compute_hydrostatics(self, *, rho=1000.0, g=9.81, divergence=None):
        """Compute hydrostatics of the FloatingBody.

        Parameters
        ----------
        rho : float, optional
            Density of Water. The default is 1000.
        g: float, optional
            Gravity acceleration. The default is 9.81.
        divergence : np.ndarray, optional
            Divergence of the DOFs.

        Returns
        -------
        hydrostatics : dict
            All hydrostatics values of the FloatingBody.
        """
        if self.center_of_mass is None:
            raise ValueError(f"Trying to compute hydrostatics for {self.name}, but no center of mass has been defined.\n"
                             f"Suggested solution: define a `center_of_mass` attribute for the FloatingBody {self.name}.")

        immersed_self = self.immersed_part()

        full_mesh_vertices = self.mesh.vertices
        coord_max = full_mesh_vertices.max(axis=0)
        coord_min = full_mesh_vertices.min(axis=0)
        full_length, full_breadth, depth = full_mesh_vertices.max(axis=0) - full_mesh_vertices.min(axis=0)

        vertices = immersed_self.mesh.vertices
        sub_length, sub_breadth, _ = vertices.max(axis=0) - vertices.min(axis=0)

        if abs(immersed_self.waterplane_area) > 1e-10:
            water_plane_idx = np.isclose(vertices[:,2], 0.0)
            water_plane = vertices[water_plane_idx][:,:-1]
            wl_length, wl_breadth = water_plane.max(axis=0) - water_plane.min(axis=0)
        else:
            wl_length, wl_breadth = 0.0, 0.0

        hydrostatics = {}
        hydrostatics["g"] = g
        hydrostatics["rho"] = rho
        hydrostatics["center_of_mass"] = self.center_of_mass

        hydrostatics["wet_surface_area"] = immersed_self.wet_surface_area
        hydrostatics["disp_volumes"] = immersed_self.volumes
        hydrostatics["disp_volume"] = immersed_self.volume
        hydrostatics["disp_mass"] = immersed_self.disp_mass(rho=rho)
        hydrostatics["center_of_buoyancy"] = immersed_self.center_of_buoyancy
        hydrostatics["waterplane_center"] = np.append(immersed_self.waterplane_center, 0.0)
        hydrostatics["waterplane_area"] = immersed_self.waterplane_area
        hydrostatics["transversal_metacentric_radius"] = immersed_self.transversal_metacentric_radius
        hydrostatics["longitudinal_metacentric_radius"] = immersed_self.longitudinal_metacentric_radius
        hydrostatics["transversal_metacentric_height"] = immersed_self.transversal_metacentric_height
        hydrostatics["longitudinal_metacentric_height"] = immersed_self.longitudinal_metacentric_height
        self.hydrostatic_stiffness = hydrostatics["hydrostatic_stiffness"] = immersed_self.compute_hydrostatic_stiffness(
            divergence=divergence, rho=rho, g=g)

        hydrostatics["length_overall"] = full_length
        hydrostatics["breadth_overall"] = full_breadth
        hydrostatics["depth"] = depth
        hydrostatics["draught"] = np.abs(coord_min[2])
        hydrostatics["length_at_waterline"] = wl_length
        hydrostatics["breadth_at_waterline"] = wl_breadth
        hydrostatics["length_overall_submerged"] = sub_length
        hydrostatics["breadth_overall_submerged"] = sub_breadth
        if any(dof.lower() in {"surge", "sway", "heave", "roll", "pitch", "yaw"}
               for dof in self.dofs) > 0: # If there is at least one rigid body dof:
            self.inertia_matrix = hydrostatics["inertia_matrix"] = self.compute_rigid_body_inertia(rho=rho)

        return hydrostatics


    ###################
    # Transformations #
    ###################

    def __add__(self, body_to_add: 'FloatingBody') -> 'FloatingBody':
        return self.join_bodies(body_to_add)

    def join_bodies(*bodies, name=None) -> 'FloatingBody':
        if name is None:
            name = "+".join(body.name for body in bodies)
        meshes = CollectionOfMeshes(
                [body.mesh.copy() for body in bodies],
                name=f"{name}_mesh"
                )
        if all(body.lid_mesh is None for body in bodies):
            lid_meshes = None
        else:
            lid_meshes = CollectionOfMeshes(
                    [body.lid_mesh.copy() for body in bodies if body.lid_mesh is not None],
                    name=f"{name}_lid_mesh"
                    )
        dofs = FloatingBody.combine_dofs(bodies)

        if all(body.mass is not None for body in bodies):
            new_mass = sum(body.mass for body in bodies)
        else:
            new_mass = None

        if (all(body.mass is not None for body in bodies)
                and all(body.center_of_mass is not None for body in bodies)):
            new_cog = sum(body.mass*np.asarray(body.center_of_mass) for body in bodies)/new_mass
        else:
            new_cog = None

        joined_bodies = FloatingBody(
            mesh=meshes, lid_mesh=lid_meshes, dofs=dofs,
            mass=new_mass, center_of_mass=new_cog, name=name
            )

        for matrix_name in ["inertia_matrix", "hydrostatic_stiffness"]:
            if all(hasattr(body, matrix_name) for body in bodies):
                from scipy.linalg import block_diag
                setattr(joined_bodies, matrix_name, joined_bodies.add_dofs_labels_to_matrix(
                        block_diag(*[getattr(body, matrix_name) for body in bodies])
                        ))

        return joined_bodies

    @staticmethod
    def combine_dofs(bodies) -> dict:
        """Combine the degrees of freedom of several bodies."""
        for body in bodies:
            body._check_dofs_shape_consistency()
        dofs = {}
        cum_nb_faces = accumulate(chain([0], (body.mesh.nb_faces for body in bodies)))
        total_nb_faces = sum(body.mesh.nb_faces for body in bodies)
        for body, nbf in zip(bodies, cum_nb_faces):
            # nbf is the cumulative number of faces of the previous subbodies,
            # that is the offset of the indices of the faces of the current body.
            for name, dof in body.dofs.items():
                new_dof = np.zeros((total_nb_faces, 3))
                new_dof[nbf:nbf+len(dof), :] = dof
                if '__' not in name:
                    new_dof_name = '__'.join([body.name, name])
                else:
                    # The body is probably a combination of bodies already.
                    # So for the associativity of the + operation,
                    # it is better to keep the same name.
                    new_dof_name = name
                dofs[new_dof_name] = new_dof
        return dofs

    def copy(self, name=None) -> 'FloatingBody':
        """Return a deep copy of the body.

        Parameters
        ----------
        name : str, optional
            a name for the new copy
        """
        self._check_dofs_shape_consistency()

        new_body = copy.deepcopy(self)
        if name is None:
            new_body.name = f"copy_of_{self.name}"
            LOG.debug(f"Copy {self.name}.")
        else:
            new_body.name = name
            LOG.debug(f"Copy {self.name} under the name {name}.")
        return new_body

    def assemble_regular_array(self, distance, nb_bodies):
        """Create an regular array of identical bodies.

        Parameters
        ----------
        distance : float
            Center-to-center distance between objects in the array
        nb_bodies : couple of ints
            Number of objects in the x and y directions.

        Returns
        -------
        FloatingBody
        """
        bodies = (self.translated((i*distance, j*distance, 0), name=f"{i}_{j}") for j in range(nb_bodies[1]) for i in range(nb_bodies[0]))
        array = FloatingBody.join_bodies(*bodies)
        array.mesh = build_regular_array_of_meshes(self.mesh, distance, nb_bodies)
        array.name = f"array_of_{self.name}"
        return array

    def assemble_arbitrary_array(self, locations:np.ndarray):

        if not isinstance(locations, np.ndarray):
            raise TypeError('locations must be of type np.ndarray')
        assert locations.shape[1] == 2, 'locations must be of shape nx2, received {:}'.format(locations.shape)

        fb_list = []
        for idx, li in enumerate(locations):
            fb1 = self.copy()
            fb1.translate(np.append(li,0))
            fb1.name = 'arbitrary_array_body{:02d}'.format(idx)
            fb_list.append(fb1)

        arbitrary_array = fb_list[0].join_bodies(*fb_list[1:])

        return arbitrary_array

    def extract_faces(self, id_faces_to_extract, return_index=False):
        """Create a new FloatingBody by extracting some faces from the mesh.
        The dofs evolve accordingly.
        The lid_mesh, center_of_mass, mass and hydrostatics data are discarded.
        """
        if isinstance(self.mesh, CollectionOfMeshes):
            raise NotImplementedError  # TODO

        if return_index:
            new_mesh, id_v = self.mesh.extract_faces(id_faces_to_extract, return_index)
        else:
            new_mesh = self.mesh.extract_faces(id_faces_to_extract, return_index)
        new_body = FloatingBody(new_mesh)
        LOG.info(f"Extract floating body from {self.name}.")

        new_body.dofs = {}
        for name, dof in self.dofs.items():
            new_body.dofs[name] = dof[id_faces_to_extract, :]

        if return_index:
            return new_body, id_v
        else:
            return new_body

    def sliced_by_plane(self, plane):
        """Return the same body, but replace the mesh by a set of two meshes
        corresponding to each sides of the plane."""
        return FloatingBody(mesh=self.mesh.sliced_by_plane(plane),
                            lid_mesh=self.lid_mesh.sliced_by_plane(plane)
                                        if self.lid_mesh is not None else None,
                            dofs=self.dofs, name=self.name)

    def minced(self, nb_slices=(8, 8, 4)):
        """Experimental method decomposing the mesh as a hierarchical structure.

        Parameters
        ----------
        nb_slices: Tuple[int, int, int]
            The number of slices in each of the x, y and z directions.
            Only powers of 2 are supported at the moment.

        Returns
        -------
        FloatingBody
        """
        minced_body = self.copy()

        # Extreme points of the mesh in each directions.
        x_min, x_max, y_min, y_max, z_min, z_max = self.mesh.axis_aligned_bbox
        sizes = [(x_min, x_max), (y_min, y_max), (z_min, z_max)]

        directions = [np.array(d) for d in [(1, 0, 0), (0, 1, 0), (0, 0, 1)]]

        def _slice_positions_at_depth(i):
            """Helper function.

            Returns a list of floats as follows:
            i=1 -> [1/2]
            i=2 -> [1/4, 3/4]
            i=3 -> [1/8, 3/8, 5/8, 7/8]
                   ...
            """
            denominator = 2**i
            return [numerator/denominator for numerator in range(1, denominator, 2)]

        # GENERATE ALL THE PLANES THAT WILL BE USED TO MINCE THE MESH
        planes = []
        for direction, nb_slices_in_dir, (min_coord, max_coord) in zip(directions, nb_slices, sizes):
            planes_in_dir = []

            depth_of_treelike_structure = int(np.log2(nb_slices_in_dir))
            for i_depth in range(1, depth_of_treelike_structure+1):
                planes_in_dir_at_depth = []
                for relative_position in _slice_positions_at_depth(i_depth):
                    slice_position = (min_coord + relative_position*(max_coord-min_coord))*direction
                    plane = Plane(normal=direction, point=slice_position)
                    planes_in_dir_at_depth.append(plane)
                planes_in_dir.append(planes_in_dir_at_depth)
            planes.append(planes_in_dir)

        # SLICE THE MESH
        intermingled_x_y_z = chain.from_iterable(zip_longest(*planes))
        for planes in intermingled_x_y_z:
            if planes is not None:
                for plane in planes:
                    minced_body = minced_body.sliced_by_plane(plane)
        return minced_body

    @inplace_transformation
    def mirror(self, plane):
        self.mesh.mirror(plane)
        if self.lid_mesh is not None:
            self.lid_mesh.mirror(plane)
        self._evaluate_full_mesh()
        for dof in self.dofs:
            self.dofs[dof] -= 2 * np.outer(np.dot(self.dofs[dof], plane.normal), plane.normal)
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                point = np.array(self.__dict__[point_attr])
                shift = - 2 * (np.dot(point, plane.normal) - plane.c) * plane.normal
                self.__dict__[point_attr] = point + shift
        return self

    @inplace_transformation
    def translate(self, vector, *args, **kwargs):
        self.mesh.translate(vector, *args, **kwargs)
        if self.lid_mesh is not None:
            self.lid_mesh.translate(vector, *args, **kwargs)
        self._evaluate_full_mesh()
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                self.__dict__[point_attr] = np.array(self.__dict__[point_attr]) + vector
        return self

    @inplace_transformation
    def rotate(self, axis, angle):
        self.mesh.rotate(axis, angle)
        if self.lid_mesh is not None:
            self.lid_mesh.rotate(axis, angle)
        self._evaluate_full_mesh()
        for point_attr in ('geometric_center', 'rotation_center', 'center_of_mass'):
            if point_attr in self.__dict__ and self.__dict__[point_attr] is not None:
                self.__dict__[point_attr] = axis.rotate_points([self.__dict__[point_attr]], angle)[0, :]
        for dof in self.dofs:
            self.dofs[dof] = axis.rotate_vectors(self.dofs[dof], angle)
        return self

    @inplace_transformation
    def clip(self, plane):
        self._check_dofs_shape_consistency()

        # Clip mesh
        LOG.info(f"Clipping {self.name} with respect to {plane}")
        self.mesh.clip(plane)
        if self.lid_mesh is not None:
            self.lid_mesh.clip(plane)
            if self.lid_mesh.nb_faces == 0:
                LOG.warning("Lid mesh %s is empty after clipping. The lid mesh is removed.", self.lid_mesh)
                self.lid_mesh = None
        self._evaluate_full_mesh()

        # Clip dofs
        ids = self.mesh._clipping_data['faces_ids']
        for dof in self.dofs:
            if len(ids) > 0:
                self.dofs[dof] = np.array(self.dofs[dof])[ids]
            else:
                self.dofs[dof] = np.empty((0, 3))
        return self


    #############
    #  Display  #
    #############

    def __short_str__(self):
        return (f"{self.__class__.__name__}(..., name=\"{self.name}\")")

    def _optional_params_str(self):
        items = []
        if self.mass is not None: items.append(f"mass={self.mass}, ")
        if self.center_of_mass is not None: items.append(f"center_of_mass={self.center_of_mass}, ")
        return ''.join(items)

    def __str__(self):
        short_dofs = '{' + ', '.join('"{}": ...'.format(d) for d in self.dofs) + '}'

        if self.lid_mesh is not None:
            lid_mesh_str = self.lid_mesh.__short_str__()
        else:
            lid_mesh_str = str(None)

        return (f"{self.__class__.__name__}(mesh={self.mesh.__short_str__()}, lid_mesh={lid_mesh_str}, "
                f"dofs={short_dofs}, {self._optional_params_str()}name=\"{self.name}\")")

    def __repr__(self):
        short_dofs = '{' + ', '.join('"{}": ...'.format(d) for d in self.dofs) + '}'

        if self.lid_mesh is not None:
            lid_mesh_str = str(self.lid_mesh)
        else:
            lid_mesh_str = str(None)

        return (f"{self.__class__.__name__}(mesh={str(self.mesh)}, lid_mesh={lid_mesh_str}, "
                f"dofs={short_dofs}, {self._optional_params_str()}name=\"{self.name}\")")

    def _repr_pretty_(self, p, cycle):
        p.text(self.__str__())

    def __rich_repr__(self):
        class DofWithShortRepr:
            def __repr__(self):
                return '...'
        yield "mesh", self.mesh
        yield "lid_mesh", self.lid_mesh
        yield "dofs", {d: DofWithShortRepr() for d in self.dofs}
        if self.mass is not None:
            yield "mass", self.mass, None
        if self.center_of_mass is not None:
            yield "center_of_mass", tuple(self.center_of_mass)
        yield "name", self.name

    def show(self, **kwargs):
        from capytaine.ui.vtk.body_viewer import FloatingBodyViewer
        viewer = FloatingBodyViewer()
        viewer.add_body(self, **kwargs)
        viewer.show()
        viewer.finalize()

    def show_matplotlib(self, *args, **kwargs):
        return self.mesh.show_matplotlib(*args, **kwargs)

    def animate(self, motion, *args, **kwargs):
        """Display a motion as a 3D animation.

        Parameters
        ==========
        motion: dict or pd.Series or str
            A dict or series mapping the name of the dofs to its amplitude.
            If a single string is passed, it is assumed to be the name of a dof
            and this dof with a unit amplitude will be displayed.
        """
        from capytaine.ui.vtk.animation import Animation
        if isinstance(motion, str):
            motion = {motion: 1.0}
        elif isinstance(motion, xr.DataArray):
            motion = {k: motion.sel(radiating_dof=k).data for k in motion.coords["radiating_dof"].data}

        if any(dof not in self.dofs for dof in motion):
            missing_dofs = set(motion.keys()) - set(self.dofs.keys())
            raise ValueError(f"Trying to animate the body {self.name} using dof(s) {missing_dofs}, but no dof of this name is defined for {self.name}.")

        animation = Animation(*args, **kwargs)
        animation._add_actor(self.mesh.merged(), faces_motion=sum(motion[dof_name] * dof for dof_name, dof in self.dofs.items() if dof_name in motion))
        return animation

    @property
    def minimal_computable_wavelength(self):
        """For accuracy of the resolution, wavelength should not be smaller than this value."""
        if self.lid_mesh is not None:
            return max(8*self.mesh.faces_radiuses.max(), 8*self.lid_mesh.faces_radiuses.max())
        else:
            return 8*self.mesh.faces_radiuses.max()

    @lru_cache
    def first_irregular_frequency_estimate(self, *, g=9.81):
        r"""Estimates the angular frequency of the lowest irregular
        frequency.
        This is based on the formula for the lowest irregular frequency of a
        parallelepiped of size :math:`L \times B` and draft :math:`H`:

        .. math::
            \omega = \sqrt{
                        \frac{\pi g \sqrt{\frac{1}{B^2} + \frac{1}{L^2}}}
                             {\tanh\left(\pi H \sqrt{\frac{1}{B^2} + \frac{1}{L^2}} \right)}
                     }

        The formula is applied to all shapes to get an estimate that is usually
        conservative.
        The definition of a lid (supposed to be fully covering and horizontal)
        is taken into account.
        """
        if self.lid_mesh is None:
            draft = abs(self.mesh.vertices[:, 2].min())
        else:
            draft = abs(self.lid_mesh.vertices[:, 2].min())
            if draft < 1e-6:
                return np.inf

        # Look for the x and y span of each components (e.g. for multibody) and
        # keep the one causing the lowest irregular frequency.
        # The draft is supposed to be same for all components.
        omega = np.inf
        for comp in connected_components(self.mesh):
            for ccomp in connected_components_of_waterline(comp):
                x_span = ccomp.vertices[:, 0].max() - ccomp.vertices[:, 0].min()
                y_span = ccomp.vertices[:, 1].max() - ccomp.vertices[:, 1].min()
                p = np.hypot(1/x_span, 1/y_span)
                omega_comp = np.sqrt(np.pi*g*p/(np.tanh(np.pi*draft*p)))
                omega = min(omega, omega_comp)
        return omega

    def cluster_bodies(*bodies, name=None):
        """
        Builds a hierarchical clustering from a group of bodies

        Parameters
        ----------
        bodies: list
            a list of bodies
        name: str, optional
            a name for the new body

        Returns
        -------
        FloatingBody
            Array built from the provided bodies
        """
        from scipy.cluster.hierarchy import linkage
        nb_buoys = len(bodies)

        if any(body.center_of_buoyancy is None for body in bodies):
            raise ValueError("The center of buoyancy of each body needs to be known for clustering")
        buoys_positions = np.stack([body.center_of_buoyancy for body in bodies])[:,:2]

        ln_matrix = linkage(buoys_positions, method='centroid', metric='euclidean')

        node_list = list(bodies)  # list of nodes of the tree: the first nodes are single bodies

        # Join the bodies, with an ordering consistent with the dendrogram.
        # Done by reading the linkage matrix: its i-th row contains the labels
        # of the two nodes that are merged to form the (n + i)-th node
        for ii in range(len(ln_matrix)):
            node_tag = ii + nb_buoys # the first nb_buoys tags are already taken
            merge_left = int(ln_matrix[ii,0])
            merge_right = int(ln_matrix[ii,1])
            # The new node is the parent of merge_left and merge_right
            new_node_ls = [node_list[merge_left], node_list[merge_right]]
            new_node = FloatingBody.join_bodies(*new_node_ls, name='node_{:d}'.format(node_tag))
            node_list.append(new_node)

        # The last node is the parent of all others
        all_buoys = new_node

        if name is not None:
            all_buoys.name = name

        return all_buoys



import os
import shutil
import pytest
import numpy as np
import capytaine as cpt


def test_dataset_from_wamit_via_bemio():
    bemio = pytest.importorskip("bemio.io.wamit", reason="Bemio not installed, test skipped.")
    current_file_path = os.path.dirname(os.path.abspath(__file__))
    out_file = os.path.join(current_file_path, "Bemio_verification_cases", "sphere.out")
    bemio_data = bemio.read(out_file)

    new_dataset = cpt.assemble_dataset(bemio_data)
    assert (np.moveaxis(bemio_data.body[0].am.all, 2, 0) * bemio_data.body[0].rho == \
        new_dataset['added_mass'].values).all()


def test_dataset_from_nemoh_via_bemio(tmp_path):
    bemio = pytest.importorskip("bemio.io.nemoh", reason="Bemio not installed, test skipped.")
    current_file_path = os.path.dirname(os.path.abspath(__file__))
    nemoh_data = os.path.join(current_file_path, "Nemoh_verification_cases", "Cylinder")
    shutil.copytree(nemoh_data, tmp_path, dirs_exist_ok=True)
    shutil.move(tmp_path / "reference_results", tmp_path / "Results")
    bemio_data = bemio.read(tmp_path)
    cpt.assemble_dataset(bemio_data)



"""This module is used for the handling of zero and infinite frequencies.
In this cases, the magnitudes that the solver has to manipulate are in the form of ω times a non-zero term.
Instead of evaluating this multiplication as zero of infinity, we keep it symbolic using the class defined here.

The frequency can be provided to the solver as something like
`SymbolicMultiplication("0", 1.0)` (that is zero) and the solver will return an
output of the form `SymbolicMultiplication("0", np.array(...))`
(that is also actually zero, except we may be intested in the non-zero array).
"""

import numpy as np
from functools import wraps, total_ordering

class SymbolicMultiplication:
    def __init__(self, symbol, value=1.0):
        self.symbol = symbol
        self.value = value
        if hasattr(value, "shape"):
            self.shape = value.shape  # When wrapping Numpy arrays

    def __format__(self, format_spec):
        return f"{self.symbol}×{self.value.__format__(format_spec)}"

    __array_priority__ = 1.0

    def __array_function__(self, func, types, *args, **kwargs):
        if func in {np.real, np.imag, np.sum}:
            return SymbolicMultiplication(self.symbol, func(self.value))
        else:
            return NotImplemented

    def __str__(self):
        return f"{self.symbol}×{self.value}"

    def __repr__(self):
        return f"SymbolicMultiplication(\"{self.symbol}\", {repr(self.value)})"

    def __add__(self, x):
        return self._concretize() + x

    def __radd__(self, x):
        return x + self._concretize()

    def __neg__(self):
        return SymbolicMultiplication(self.symbol, -self.value)

    def __mul__(self, x):
        return SymbolicMultiplication(self.symbol, self.value * x)

    def __rmul__(self, x):
        return SymbolicMultiplication(self.symbol, x * self.value)

    def __pow__(self, n):
        if n == 2:
            return self * self
        else:
            raise NotImplementedError

    def __truediv__(self, x):
        if hasattr(x, 'symbol') and self.symbol == x.symbol:
            return self.value / x.value
        else:
            return SymbolicMultiplication(self.symbol, self.value / x)

    def __rtruediv__(self, x):
        if hasattr(x, 'symbol') and self.symbol == x.symbol:
            return x.value / self.value
        elif self.symbol == "0":
            return SymbolicMultiplication("∞", x/self.value)
        elif self.symbol == "∞":
            return SymbolicMultiplication("0", x/self.value)
        else:
            raise NotImplementedError

    def __matmul__(self, x):
        return SymbolicMultiplication(self.symbol, self.value @ x)

    def __rmatmul__(self, x):
        return SymbolicMultiplication(self.symbol, x @ self.value)

    def __getitem__(self, item):
        return SymbolicMultiplication(self.symbol, self.value[item])

    def __setitem__(self, item, val):
        if isinstance(val, SymbolicMultiplication) and self.symbol == val.symbol:
            self.value.__setitem__(item, val.value)
        else:
            raise NotImplementedError

    def __lt__(self, x):
        return self._concretize() < x

    def __le__(self, x):
        return self._concretize() <= x

    def __eq__(self, x):
        return self._concretize() == x

    def __ge__(self, x):
        return self._concretize() >= x

    def __gt__(self, x):
        return self._concretize() > x

    def __hash__(self):
        return hash((self.symbol, self.value))

    def _concretize(self):
        if isinstance(self.value, np.ndarray):
            if self.symbol == "0":
                return np.zeros_like(self.value)
            elif self.symbol == "∞":
                return np.full_like(self.value, np.inf)
        else:
            return float(self)

    def __float__(self):
        if self.symbol == "0":
            return 0.0 * float(self.value)
        elif self.symbol == "∞":
            return np.inf * float(self.value)
        else:
            raise NotImplementedError

    def reshape(self, *args):
        return SymbolicMultiplication(self.symbol, self.value.reshape(*args))

    def sum(self, *args, **kwargs):
        return SymbolicMultiplication(self.symbol, self.value.sum(*args, **kwargs))

    @property
    def T(self):
        return SymbolicMultiplication(self.symbol, self.value.T)


def supporting_symbolic_multiplication(f):
    """
    When this decorator is applied to a function, this function can now take
    as input a `SymbolicMultiplication` object. The function is applied on the
    `value` part of the `SymbolicMultiplication` without modifying the
    `symbol`.
    """
    @wraps(f)
    def wrapped_f(a, x):
        if hasattr(x, 'symbol'):
            return SymbolicMultiplication(x.symbol, f(a, x.value))
        else:
            return f(a, x)
    return wrapped_f



import numpy as np
import capytaine as cpt

cpt.set_logging('INFO')

radius = 12.5
draught = 37.5
n_radius = 15
n_theta = 25
n_z = 30

# Define the range of frequencies to solve
omega_range = np.linspace(1.5, 2.25, 50)

solver = cpt.BEMSolver()

def compute_excitation_force(body, omega_range):
    problems = [
        cpt.DiffractionProblem(body=body, omega=omega)
        for omega in omega_range
    ]
    results = solver.solve_all(problems)
    data = cpt.assemble_dataset(results)
    return data['excitation_force']


# Initialize floating body by generating a geometric mesh
cylinder_mesh = cpt.mesh_vertical_cylinder(
    length=draught*2, radius=radius, center=(0, 0, 0), faces_max_radius=1.0,
    ).immersed_part()

### Body Without Lid (Body A)
body_1 = cpt.FloatingBody(mesh=cylinder_mesh)
body_1.add_translation_dof(name="Surge")

### Body With Lid at z = -3.699 (Body B)
lid_mesh = cylinder_mesh.generate_lid(z=-3.699)
body_2 = cpt.FloatingBody(mesh=cylinder_mesh, lid_mesh=lid_mesh)
body_2.add_translation_dof(name="Surge")

### Body With Lid at z = auto (Body C)
lid_mesh = cylinder_mesh.generate_lid(z=cylinder_mesh.lowest_lid_position(omega_max=omega_range.max()))
body_3 = cpt.FloatingBody(mesh=cylinder_mesh, lid_mesh=lid_mesh)
body_3.add_translation_dof(name="Surge")

force_1 = compute_excitation_force(body_1, omega_range)
force_2 = compute_excitation_force(body_2, omega_range)
force_3 = compute_excitation_force(body_3, omega_range)

# Plot the added mass of each dofs as a function of the frequency
import matplotlib.pyplot as plt
plt.figure()
plt.plot(
    omega_range,
    np.abs(force_1.sel(influenced_dof='Surge')),
    marker='+',
    label='no lid'
)
plt.plot(
    omega_range,
    np.abs(force_2.sel(influenced_dof='Surge')),
    marker='*',
    label=r'$z = -3.699$m'
)
plt.plot(
    omega_range,
    np.abs(force_3.sel(influenced_dof='Surge')),
    marker='x',
    label='auto'
)

plt.xlabel(r'$\omega$')
plt.ylabel(r'$F_1 (abs)$')
plt.legend()
plt.tight_layout()
plt.show()



import numpy as np
import pytest

import capytaine as cpt


@pytest.mark.parametrize("z_center", [-10.0, 0.0, 10.0])
def test_analytical_solution(z_center):
    radius = 1.0
    mesh = cpt.mesh_sphere(center=(0, 0, z_center), radius=radius, resolution=(10, 10))
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    pb = cpt.RadiationProblem(body=body, free_surface=np.inf, radiating_dof="Surge")
    solver = cpt.BEMSolver(method="direct")
    res = solver.solve(pb)
    assert np.allclose(2/3*np.pi*pb.rho*radius**3, res.forces["Surge"], rtol=1e-2)


def test_translation_invariance_of_no_free_surface_case():
    def force_on_body(z):
        mesh = cpt.mesh_parallelepiped(center=(0, 0, z))
        body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)))
        pb = cpt.RadiationProblem(body=body, free_surface=np.inf, water_depth=np.inf, radiating_dof="Surge")
        solver = cpt.BEMSolver(method="direct")
        res = solver.solve(pb)
        return res.force["Surge"]
    assert np.isclose(force_on_body(0.0), force_on_body(-1.0))
    assert np.isclose(force_on_body(0.0), force_on_body(1.0))



import re
import logging
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
import pytest

import capytaine as cpt
from capytaine.io.wamit import export_to_wamit


@pytest.fixture
def full_dataset():
    mesh = cpt.mesh_sphere(resolution=(4, 4))
    dofs = cpt.rigid_body_dofs(rotation_center=(0, 0, 0))
    full_body = cpt.FloatingBody(mesh, dofs)
    full_body.center_of_mass = np.copy(full_body.mesh.center_of_mass_of_nodes)
    immersed_body = full_body.immersed_part()
    immersed_body.compute_hydrostatics()
    test_matrix = xr.Dataset(
        {
            "omega": [0.0, 1.0, np.inf],
            "wave_direction": [0.0],
            "radiating_dof": list(immersed_body.dofs),
            "water_depth": [np.inf],
            "rho": [1025],
        }
    )
    solver = cpt.BEMSolver()
    dataset = solver.fill_dataset(test_matrix, immersed_body)

    return dataset


@pytest.fixture
def dataset_with_multiple_rho():
    mesh = cpt.mesh_sphere(resolution=(4, 4))
    dofs = cpt.rigid_body_dofs(rotation_center=(0, 0, 0))
    full_body = cpt.FloatingBody(mesh, dofs)
    full_body.center_of_mass = np.copy(full_body.mesh.center_of_mass_of_nodes)
    immersed_body = full_body.immersed_part()

    # Define a test matrix with multiple rho
    test_matrix = xr.Dataset(
        {
            "omega": ("omega", [1.0]),
            "wave_direction": ("wave_direction", [0.0]),
            "radiating_dof": ("radiating_dof", list(immersed_body.dofs)),
            "water_depth": ("water_depth", [np.inf]),
            "rho": ("rho", [1020.0, 1025.0]),  # ➜ should lead to an error
        }
    )

    solver = cpt.BEMSolver()
    dataset = solver.fill_dataset(test_matrix, immersed_body)
    return dataset


def test_export_wamit_and_import_force(full_dataset, tmpdir):
    """Export to WAMIT (.1 and .3 files), then reload and check that the frequency matches."""
    dataset = full_dataset.sel(omega=[1.0])
    problem_name = str(tmpdir / "boat_200_unique")
    export_to_wamit(dataset, problem_name=problem_name, exports=("1", "3"))
    wamit_1_path = tmpdir / "boat_200_unique.1"
    wamit_3_path = tmpdir / "boat_200_unique.3"
    assert wamit_1_path.exists(), f"File {wamit_1_path} was not generated."
    assert wamit_3_path.exists(), f"File {wamit_3_path} was not generated."

    # Now import the .1 file and check the frequency
    imported_df = pd.read_csv(wamit_1_path, sep=r"\s+", header=None)
    # The first column is PERIOD
    imported_period = float(imported_df.iloc[0, 0])
    ref_period = float(dataset.period)
    assert np.isclose(
        imported_period, ref_period
    ), f"Imported period {imported_period} does not match expected {ref_period}"

    # Now import the .3 file and check the frequency
    imported_df = pd.read_csv(wamit_3_path, sep=r"\s+", header=None)
    # The first column is omega
    imported_period = float(imported_df.iloc[0, 0])
    assert np.isclose(
        imported_period, ref_period
    ), f"Imported period {imported_period} does not match expected {ref_period}"


def test_export_wamit_omega_zero(full_dataset, tmpdir):
    """
    Test a simulation with omega=0, export to WAMIT, and check the .1 file:
    - The file must exist
    - All data lines must have 4 fields
    - The number of data lines must be len(omega) * 36 (here: 36)
    """
    dataset = full_dataset.sel(omega=[0.0])
    export_to_wamit(
        dataset, problem_name=str(tmpdir / "boat_200_omega0"), exports=("1",)
    )
    wamit_1_path = tmpdir / "boat_200_omega0.1"
    assert wamit_1_path.exists(), f"File {wamit_1_path} was not generated."
    with open(wamit_1_path, "r", encoding="utf-8") as f:
        lines = [line.strip() for line in f if line.strip()]
    data_lines = [
        line
        for line in lines
        if re.match(r"^\s*[-+eE0-9.]+\s+\d+\s+\d+\s+[-+eE0-9.]+$", line)
    ]
    assert data_lines, "No data lines found in the .1 file for omega=0."
    for line in data_lines:
        fields = re.split(r"\s+", line)
        assert (
            len(fields) == 4
        ), f"For omega=0, each line in the .1 file must have 4 fields: {fields}"
    expected_lines = len(dataset["omega"]) * 36
    assert (
        len(data_lines) == expected_lines
    ), f"Expected {expected_lines} data lines, found {len(data_lines)}."


def test_export_wamit_omega_inf(full_dataset, tmpdir):
    """
    Test a simulation with omega=inf, export to WAMIT, and check the .1 file:
    - The file must exist
    - All data lines must have 4 fields
    - The number of data lines must be len(omega) * 36 (here: 36)
    """
    dataset = full_dataset.sel(omega=[np.inf])
    export_to_wamit(
        dataset, problem_name=str(tmpdir / "boat_200_omegaINF"), exports=("1",)
    )
    wamit_1_path = tmpdir / "boat_200_omegaINF.1"
    assert wamit_1_path.exists(), f"File {wamit_1_path} was not generated."
    with open(wamit_1_path, "r", encoding="utf-8") as f:
        lines = [line.strip() for line in f if line.strip()]
    data_lines = [
        line
        for line in lines
        if re.match(r"^\s*[-+eE0-9.]+\s+\d+\s+\d+\s+[-+eE0-9.]+$", line)
    ]
    assert data_lines, "No data lines found in the .1 file for omega=inf."
    for line in data_lines:
        fields = re.split(r"\s+", line)
        assert (
            len(fields) == 4
        ), f"For omega=inf, each line in the .1 file must have 4 fields: {fields}"
    expected_lines = len(dataset["omega"]) * 36
    assert (
        len(data_lines) == expected_lines
    ), f"Expected {expected_lines} data lines, found {len(data_lines)}."


def test_export_wamit_hydrostatic(full_dataset, tmpdir):
    """
    Test a simulation with multiple classic omega values, export hydrostatics to WAMIT, and check the .hst file:
    - The file must exist
    - All data lines must have 3 fields
    - The number of data lines must be 36 (6x6 matrix, possibly sparse)
    """
    export_to_wamit(
        full_dataset, problem_name=str(tmpdir / "boat_200_omega"), exports=("hst",)
    )
    wamit_hst_path = tmpdir / "boat_200_omega.hst"
    assert wamit_hst_path.exists(), f"File {wamit_hst_path} was not generated."
    with open(wamit_hst_path, "r", encoding="utf-8") as f:
        lines = [line.strip() for line in f if line.strip()]
    data_lines = [
        line for line in lines if re.match(r"^\s*\d+\s+\d+\s+[-+eE0-9.]+$", line)
    ]
    assert data_lines, "No data lines found in the .hst file for omega classic."
    for line in data_lines:
        fields = re.split(r"\s+", line)
        assert (
            len(fields) == 3
        ), f"For hydrostatics, each line in the .hst file must have 3 fields: {fields}"
    expected_lines = 36  # 6 dofs * 6 directions = 36
    assert (
        len(data_lines) == expected_lines
    ), f"Expected {expected_lines} data lines, found {len(data_lines)}."


def test_warning_for_wrong_format(full_dataset, tmpdir, caplog):
    """Test the warning when trying to export excitation force but they are not in the dataset"""
    radiation_only_ds = full_dataset[["added_mass", "radiation_damping"]]
    problem_name = str(tmpdir / "radiation_data")
    with caplog.at_level(logging.WARNING):
        export_to_wamit(radiation_only_ds, problem_name=problem_name, exports=["3"])
    assert "Missing field 'excitation_force'" in caplog.text
    assert not (tmpdir / "radiation_data.3").exists()


def compare_wamit_files(file1, file2, atol=1e-14):
    """Compare two WAMIT output files by parsing numerical values.

    Args:
        file1 (str): Path to the first WAMIT file.
        file2 (str): Path to the second WAMIT file.
        atol (float): Absolute tolerance for numerical comparison.

    Raises:
        AssertionError: If line counts differ or any value mismatches beyond tolerance.
    """
    with open(file1, "r", encoding="utf-8") as f:
        lines1 = [line.strip() for line in f if line.strip()]
    with open(file2, "r", encoding="utf-8") as f:
        lines2 = [line.strip() for line in f if line.strip()]

    assert len(lines1) == len(
        lines2
    ), f"Line count mismatch: {len(lines1)} vs {len(lines2)}"

    for i, (l1, l2) in enumerate(zip(lines1, lines2)):
        try:
            vals1 = [float(s) for s in re.split(r"\s+", l1)]
            vals2 = [float(s) for s in re.split(r"\s+", l2)]
        except ValueError:
            raise AssertionError(f"Line {i} contains non-numeric data:\n{l1}\n{l2}")

        assert len(vals1) == len(
            vals2
        ), f"Line {i} has different number of columns: {vals1} vs {vals2}"

        def condition(vals1, vals2, atol):
            return not np.allclose(vals1, vals2, atol=atol)

        if condition(vals1, vals2, atol):
            diffs = [
                f"{v1} ≠ {v2}"
                for v1, v2 in zip(vals1, vals2)
                if condition([v1], [v2], atol)
            ]
            raise AssertionError(
                f"Line {i} values differ beyond tolerance: {l1} // {l2} -- Differences: {', '.join(diffs)}"
            )


rng = np.random.default_rng(seed=42)
omega_values = rng.uniform(0.5, 5.0, size=3)


def test_export_wamit_fails_with_multiple_rho(dataset_with_multiple_rho, tmpdir):
    """Test that export fails when multiple rho values are present."""
    problem_name = tmpdir / "boat_invalid_rho"

    with pytest.raises(ValueError, match="only one value.*rho"):
        export_to_wamit(dataset_with_multiple_rho, problem_name=str(problem_name))


@pytest.mark.parametrize("omega_val", omega_values)
@pytest.mark.parametrize("export_type", ["hst", "1", "3"])
def test_export_wamit_frequency_axis_representations(export_type, omega_val, tmpdir):
    """
    Run Capytaine simulation with a single omega, then:
    - create two datasets: one with omega as dimension, one with period
    - export both
    - compare the resulting .1 and .3 files
    """
    # Load mesh and run BEM simulation for 1 omega
    mesh = cpt.mesh_sphere(resolution=(4, 4))
    dofs = cpt.rigid_body_dofs(rotation_center=(0, 0, 0))
    full_body = cpt.FloatingBody(mesh, dofs)
    full_body.center_of_mass = np.copy(full_body.mesh.center_of_mass_of_nodes)
    immersed_body = full_body.immersed_part()
    if export_type == "hst":
        immersed_body.compute_hydrostatics()

    test_matrix_omega = xr.Dataset(
        {
            "omega": np.asarray(omega_val),
            "wave_direction": [0.0],
            "radiating_dof": list(immersed_body.dofs),
            "water_depth": [np.inf],
            "rho": [1025],
        }
    )

    solver = cpt.BEMSolver()
    ds_omega = solver.fill_dataset(test_matrix_omega, immersed_body)

    # Create period-based version
    ds_period = ds_omega.swap_dims({"omega": "period"})
    ds_wavenumber = ds_period.swap_dims({"period": "wavenumber"})
    ds_wavelength = ds_wavenumber.swap_dims({"wavenumber": "wavelength"})

    # Export all
    tmpdir = Path(tmpdir)
    omega_name = tmpdir / f"omega_export_{export_type}"
    period_name = tmpdir / f"period_export_{export_type}"
    wavenumber_name = tmpdir / f"wavenumber_export_{export_type}"
    wavelength_name = tmpdir / f"wavelength_export_{export_type}"

    export_to_wamit(ds_omega, problem_name=str(omega_name), exports=[export_type])
    export_to_wamit(ds_period, problem_name=str(period_name), exports=[export_type])
    export_to_wamit(
        ds_wavenumber, problem_name=str(wavenumber_name), exports=[export_type]
    )
    export_to_wamit(
        ds_wavelength, problem_name=str(wavelength_name), exports=[export_type]
    )

    f_omega = omega_name.with_name(f"{omega_name.name}.{export_type}")
    f_period = period_name.with_name(f"{period_name.name}.{export_type}")
    f_wavenumber = wavenumber_name.with_name(f"{wavenumber_name.name}.{export_type}")
    f_wavelength = wavelength_name.with_name(f"{wavelength_name.name}.{export_type}")

    assert f_omega.exists(), f"File {f_omega} not generated"
    assert f_period.exists(), f"File {f_period} not generated"
    assert f_wavenumber.exists(), f"File {f_wavenumber} not generated"
    assert f_wavelength.exists(), f"File {f_wavelength} not generated"

    compare_wamit_files(f_omega, f_period)
    compare_wamit_files(f_omega, f_wavenumber)
    compare_wamit_files(f_omega, f_wavelength)



import os

import numpy as np
import pytest

import capytaine as cpt
from capytaine.io.legacy import export_hydrostatics


def test_legacy_export_hydrostatics():
    # create two cylinders & compute inertia & hydrostatic data
    cylinder1_mesh = cpt.meshes.predefined.mesh_vertical_cylinder(length=5.0, center = (1.0, 1.0, -0.5))
    cylinder1 = cpt.FloatingBody(cylinder1_mesh, center_of_mass=[1.0, 1.0, -2.0])
    cylinder1.add_all_rigid_body_dofs()
    cylinder1.compute_rigid_body_inertia()
    cylinder1.compute_hydrostatics()

    cylinder2_mesh = cpt.meshes.predefined.mesh_vertical_cylinder(length=5.0, center = (5.0, 1.0, -0.5))
    cylinder2 = cpt.FloatingBody(cylinder2_mesh, center_of_mass=[5.0, 1.0, -2.0])
    cylinder2.add_all_rigid_body_dofs()
    cylinder2.compute_rigid_body_inertia()
    cylinder2.compute_hydrostatics()

    current_file_path = os.path.dirname(os.path.abspath(__file__))
    testing_directory = os.path.join(current_file_path, 'io_legacy_cases')

    # export hydrostatics
    export_hydrostatics(os.path.join(testing_directory, 'single_body'), cylinder1)
    export_hydrostatics(os.path.join(testing_directory, 'single_body_list'), [cylinder1])
    export_hydrostatics(os.path.join(testing_directory, 'two_bodies_list'), [cylinder1, cylinder2])


    # Check single body Hydrostatics.dat
    with open(os.path.join(testing_directory, "single_body/Hydrostatics.dat"), "r") as f:
        single_body_Hydrostatics = f.read()
    with open(os.path.join(testing_directory, "reference_data/single_body/Hydrostatics.dat"), "r") as f:
        single_body_Hydrostatics_ref = f.read()
    assert single_body_Hydrostatics == single_body_Hydrostatics_ref


    # Check single body (list) Hydrostatics.dat
    with open(os.path.join(testing_directory, "single_body_list/Hydrostatics.dat"), "r") as f:
        single_body_list_Hydrostatics = f.read()
    with open(os.path.join(testing_directory, "reference_data/single_body_list/Hydrostatics.dat"), "r") as f:
        single_body_list_Hydrostatics_ref = f.read()
    assert single_body_list_Hydrostatics == single_body_list_Hydrostatics_ref


    # Check two bodies (list) Hydrostatics_0.dat
    with open(os.path.join(testing_directory, "two_bodies_list/Hydrostatics_0.dat"), "r") as f:
        two_bodies_Hydrostatics_0 = f.read()
    with open(os.path.join(testing_directory, "reference_data/two_bodies_list/Hydrostatics_0.dat"), "r") as f:
        two_bodies_Hydrostatics_0_ref = f.read()
    assert two_bodies_Hydrostatics_0 == two_bodies_Hydrostatics_0_ref

    # Check two bodies (list) Hydrostatics_1.dat
    with open(os.path.join(testing_directory, "two_bodies_list/Hydrostatics_1.dat"), "r") as f:
        two_bodies_Hydrostatics_1 = f.read()
    with open(os.path.join(testing_directory, "reference_data/two_bodies_list/Hydrostatics_1.dat"), "r") as f:
        two_bodies_Hydrostatics_1_ref = f.read()
    assert two_bodies_Hydrostatics_1 == two_bodies_Hydrostatics_1_ref


    # Check single body KH.dat
    single_body_KH = np.loadtxt(os.path.join(testing_directory, "single_body/KH.dat"))
    single_body_KH_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/single_body/KH.dat"))
    np.testing.assert_allclose(single_body_KH, single_body_KH_ref, atol=1e-6)

    # Check single body (list) KH.dat
    single_body_list_KH = np.loadtxt(os.path.join(testing_directory, "single_body_list/KH.dat"))
    single_body_list_KH_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/single_body_list/KH.dat"))
    np.testing.assert_allclose(single_body_list_KH, single_body_list_KH_ref, atol=1e-6)

    # Check two bodies (list) KH_0.dat
    two_bodies_KH_0 = np.loadtxt(os.path.join(testing_directory, "two_bodies_list/KH_0.dat"))
    two_bodies_KH_0_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/two_bodies_list/KH_0.dat"))
    np.testing.assert_allclose(two_bodies_KH_0, two_bodies_KH_0_ref, atol=1e-6)

    # Check two bodies (list) KH_1.dat
    two_bodies_KH_1 = np.loadtxt(os.path.join(testing_directory, "two_bodies_list/KH_1.dat"))
    two_bodies_KH_1_ref = np.loadtxt(os.path.join(testing_directory, "reference_data/two_bodies_list/KH_1.dat"))
    np.testing.assert_allclose(two_bodies_KH_1, two_bodies_KH_1_ref, atol=1e-6)



import numpy as np
import capytaine as cpt

# Initialize an example mesh and test case.
mesh = cpt.mesh_sphere().immersed_part()
body = cpt.FloatingBody(mesh, dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)))

pb = cpt.RadiationProblem(body=body, omega=np.inf, radiating_dof="Heave")
# Zero and infinity are accepted values for "omega", as well as for "period", "wavelength" and "wavenumber"

solver = cpt.BEMSolver()
res = solver.solve(pb)

print(f"Heave-heave added mass: {res.added_mass['Heave']:.2e}")

p = np.real(res.pressure)
# p is a custom Python object storing the pressure on each panel

i_panel = 0  # Some panel for testing

# All pressures are actually infinite
print(f"Pressure on panel {i_panel}:", float(p[i_panel]))

# However p/omega^2 is finite (hence the finite added mass)
p_over_omega2 = p/pb.omega**2
print(f"Pressure/omega^2 on panel {i_panel}: {p_over_omega2[i_panel]:.2e}")



import pytest

import numpy as np
import capytaine as cpt


def test_cache_matrices():
    """Test how the BasicMatrixEngine caches the interaction matrices."""
    mesh = cpt.mesh_sphere(radius=1.0, resolution=(4, 3)).immersed_part()
    gf = cpt.Delhommeau()
    params_1 = (mesh, mesh, 0.0, np.inf, 1.0, gf)
    params_2 = (mesh, mesh, 0.0, np.inf, 2.0, gf)

    # No cache
    engine = cpt.BasicMatrixEngine(matrix_cache_size=0)
    S, K             = engine.build_matrices(*params_1)
    S_again, K_again = engine.build_matrices(*params_1)
    assert S is not S_again
    assert K is not K_again

    # Cache
    engine = cpt.BasicMatrixEngine(matrix_cache_size=1)
    S, K                     = engine.build_matrices(*params_1)
    S_again, K_again         = engine.build_matrices(*params_1)
    _, _                     = engine.build_matrices(*params_2)
    S_once_more, K_once_more = engine.build_matrices(*params_2)
    assert S is S_again
    assert S is not S_once_more
    assert K is K_again
    assert K is not K_once_more


@pytest.mark.parametrize("method", ['indirect', 'direct'])
def test_custom_linear_solver(method):
    """Solve a simple problem with a custom linear solver."""
    sphere = cpt.FloatingBody(mesh=cpt.mesh_sphere(radius=1.0, resolution=(4, 3)).immersed_part())
    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
    problem = cpt.RadiationProblem(body=sphere, omega=1.0, water_depth=np.inf)

    reference_solver = cpt.BEMSolver(
        engine=cpt.BasicMatrixEngine(linear_solver="gmres", matrix_cache_size=0),
        method=method
    )
    reference_result = reference_solver.solve(problem)

    def my_linear_solver(A, b):
        """A dumb solver for testing."""
        return np.linalg.inv(A) @ b

    my_bem_solver = cpt.BEMSolver(
        engine=cpt.BasicMatrixEngine(linear_solver=my_linear_solver, matrix_cache_size=0),
        method=method
    )
    assert 'my_linear_solver' in my_bem_solver.exportable_settings['linear_solver']

    result = my_bem_solver.solve(problem)
    assert np.isclose(reference_result.added_masses['Surge'], result.added_masses['Surge'])


@pytest.mark.parametrize("method", ['indirect', 'direct'])
def test_custom_linear_solver_returning_wrong_shape(method):
    sphere = cpt.FloatingBody(mesh=cpt.mesh_sphere(radius=1.0, resolution=(4, 3)).immersed_part())
    sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
    problem = cpt.RadiationProblem(body=sphere, omega=1.0, water_depth=np.inf)

    def my_linear_solver(A, b):
        """A dumb solver for testing."""
        return np.linalg.inv(A) @ b.reshape(-1, 1)

    my_bem_solver = cpt.BEMSolver(
        method=method,
        engine=cpt.BasicMatrixEngine(linear_solver=my_linear_solver)
    )
    with pytest.raises(ValueError):
        my_bem_solver.solve(problem)



import numpy as np
import xarray as xr
import capytaine as cpt

from capytaine.meshes.mesh_like_protocol import MeshLike


def test_existing_classes_are_meshlike():
    mesh = cpt.mesh_sphere()
    assert isinstance(mesh, MeshLike)
    assert isinstance(cpt.CollectionOfMeshes([mesh, mesh.translated_x(0.1)]), MeshLike)

class MyMesh:
    """A set of horizontal panels at a list of depths"""
    def __init__(self, zs=[-1.0]):
        self.zs = zs
        self.vertices = np.array(
            sum(
                ([[0.0, 0.0, z], [0.0, 1.0, z], [1.0, 1.0, z], [1.0, 0.0, z]]
                for z in zs),

                []
            )
        )
        self.faces = np.array([[0, 1, 2, 3] for z in zs])
        self.faces_centers = np.array([[0.5, 0.5, z] for z in zs])
        self.faces_areas = np.array([1.0 for z in zs])
        self.faces_radiuses = np.array([np.sqrt(2) / 2 for z in zs])
        self.faces_normals = np.array([[0.0, 0.0, -1.0] for z in zs])
        self.quadrature_points = (
            self.faces_centers.reshape(len(zs), 1, 3),
            self.faces_areas,
        )
        self.nb_vertices = 4*len(zs)
        self.nb_faces = len(zs)

    def __short_str__(self):
        return str(self)

    def extract_faces(self, *args, **kwargs):
        return self

    def join_meshes(*meshes, return_masks=False):
        joined = MyMesh(sum([m.zs for m in meshes], []))
        # Concatenate the lists of vertical positions
        if not return_masks:
            return joined
        else:
            masks = [np.full((joined.nb_faces,), False) for _ in meshes]
            accumulate_shifts = 0
            for i, m in enumerate(meshes):
                masks[i][accumulate_shifts:accumulate_shifts+m.nb_faces] = True
                accumulate_shifts = accumulate_shifts + m.nb_faces
            return joined, masks

    def with_normal_vector_going_down(self, **kwargs):
        return self

def test_minimal_implementation_is_meshlike():
    assert isinstance(MyMesh(), MeshLike)

def test_minimal_implementation_solve_pb():
    mesh = MyMesh(zs=[-1.0])
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb = cpt.RadiationProblem(body=body, omega=1.0, radiating_dof="Heave")
    solver.solve(pb, method="indirect")

def test_minimal_implementation_compute_velocity():
    mesh = MyMesh(zs=[-1.0])
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb = cpt.RadiationProblem(body=body, omega=1.0, radiating_dof="Heave")
    res = solver.solve(pb, method="indirect")
    solver.compute_velocity(mesh, res)

def test_minimal_implementation_solve_pb_with_lid():
    mesh = MyMesh(zs=[-1.0])
    lid_mesh = MyMesh(zs=[0.0])
    body = cpt.FloatingBody(mesh=mesh, lid_mesh=lid_mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    pb = cpt.RadiationProblem(body=body, omega=1.0, radiating_dof="Heave")
    solver.solve(pb, method="indirect")

def test_minimal_implementation_fill_dataset():
    mesh = MyMesh(zs=[-1.0])
    body = cpt.FloatingBody(mesh=mesh, dofs=cpt.rigid_body_dofs())
    solver = cpt.BEMSolver()
    test_matrix = xr.Dataset(
        coords={"wavenumber": 1.0, "radiating_dof": ["Heave"], "wave_direction": 1.0}
    )
    solver.fill_dataset(test_matrix, body, method="direct")



import pytest
import numpy as np
from capytaine.tools.symbolic_multiplication import SymbolicMultiplication, supporting_symbolic_multiplication

def test_definition():
    zero = SymbolicMultiplication("0")
    assert zero.symbol == "0"
    assert isinstance(zero, SymbolicMultiplication)

def test_multiplication():
    zero = SymbolicMultiplication("0")
    b = 2 * zero
    assert b.value == 2 * zero.value

def test_division():
    zero = SymbolicMultiplication("0")
    b = 2 * zero
    assert b / zero == 2

def test_double_division():
    zero = SymbolicMultiplication("0")
    b = 7 * zero * zero
    assert b / (zero * zero) == 7

def test_invert_division():
    zero = SymbolicMultiplication("0")
    b = 9 * zero
    assert zero / b == pytest.approx(1/9)

def test_float():
    zero = SymbolicMultiplication("0")
    b = 4.5 * zero
    assert float(b) == pytest.approx(0.0)

def test_numpy_array():
    zero = SymbolicMultiplication("0")
    b = np.random.rand(10) * zero
    assert (b / zero).shape == (10,)

def test_numpy_array_sum():
    zero = SymbolicMultiplication("0")
    b = np.ones(10) * zero
    assert (np.sum(b) / zero) == 10

def test_numpy_matmul():
    zero = SymbolicMultiplication("0")
    b = np.random.rand(10) * zero
    A = np.random.rand(10, 10)
    c = A @ b
    assert (c/zero).shape == (10,)

def test_comparison():
    a = SymbolicMultiplication("0", np.arange(10))
    assert np.all(a == 0)

def test_setitem_mask():
    a = SymbolicMultiplication("0", np.arange(10))
    b = SymbolicMultiplication("0", 99)
    a[a.value > 5] = b
    assert np.all(a.value == np.array([0, 1, 2, 3, 4, 5, 99, 99, 99, 99]))

def test_undefined_case():
    assert np.isnan(float(SymbolicMultiplication("0", np.inf)))
    assert np.isnan(float(SymbolicMultiplication("∞", 0.0)))

def test_supporting_symbolic_multiplication():
    zero = SymbolicMultiplication("0")

    @supporting_symbolic_multiplication
    def my_linear_operator(A, x):
        return np.linalg.solve(A, x)

    b = np.random.rand(10) * zero
    A = np.random.rand(10, 10)
    c = my_linear_operator(A, b)
    assert (c/zero).shape == (10,)
